#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import copy
import json
import traceback
from collections.abc import Iterable
from typing import TYPE_CHECKING, cast

from airflow.providers.common.compat.openlineage.facet import (
    ColumnLineageDatasetFacet,
    DatasetFacet,
    ErrorMessageRunFacet,
    ExternalQueryRunFacet,
    Fields,
    InputDataset,
    InputField,
    OutputDataset,
    OutputStatisticsOutputDatasetFacet,
    SchemaDatasetFacet,
    SQLJobFacet,
)
from airflow.providers.google.cloud.openlineage.utils import (
    BIGQUERY_NAMESPACE,
    get_facets_from_bq_table,
    get_from_nullable_chain,
    get_identity_column_lineage_facet,
    get_namespace_name_from_source_uris,
    merge_column_lineage_facets,
)

if TYPE_CHECKING:
    from airflow.providers.common.compat.openlineage.facet import Dataset, RunFacet
    from airflow.providers.google.cloud.openlineage.facets import BigQueryJobRunFacet


class _BigQueryInsertJobOperatorOpenLineageMixin:
    """Mixin for BigQueryInsertJobOperator to extract OpenLineage metadata."""

    def get_openlineage_facets_on_complete(self, _):
        """
        Retrieve OpenLineage data for a completed BigQuery job.

        This method calls BigQuery API, retrieving input and output dataset info from it,
         as well as run-level statistics.

        Run facets may contain:
            - ExternalQueryRunFacet (for QUERY job type)
            - BigQueryJobRunFacet
            - ErrorMessageRunFacet (if an error occurred)

        Job facets should contain:
            - SqlJobFacet (for QUERY job type)

        Input datasets should contain:
            - SchemaDatasetFacet

        Output datasets should contain:
            - SchemaDatasetFacet
            - OutputStatisticsOutputDatasetFacet (for QUERY job type)
            - ColumnLineageDatasetFacet (for QUERY job type)
        """
        from airflow.providers.openlineage.extractors import OperatorLineage
        from airflow.providers.openlineage.sqlparser import SQLParser

        if not self.job_id:
            self.log.warning("No BigQuery job_id was found by OpenLineage.")
            return OperatorLineage()

        if not self.hook:
            # This can occur when in deferrable mode
            from airflow.providers.google.cloud.hooks.bigquery import BigQueryHook

            self.hook = BigQueryHook(
                gcp_conn_id=self.gcp_conn_id,
                impersonation_chain=self.impersonation_chain,
            )

        self.log.debug("Extracting data from bigquery job: `%s`", self.job_id)
        inputs, outputs = [], []
        run_facets: dict[str, RunFacet] = {
            "externalQuery": ExternalQueryRunFacet(externalQueryId=self.job_id, source="bigquery")
        }
        self._client = self.hook.get_client(
            project_id=self.project_id or self.hook.project_id, location=self.location
        )
        try:
            job_properties = self._client.get_job(job_id=self.job_id)._properties

            if get_from_nullable_chain(job_properties, ["status", "state"]) != "DONE":
                raise ValueError(f"Trying to extract data from running bigquery job: `{self.job_id}`")

            run_facets["bigQueryJob"] = self._get_bigquery_job_run_facet(job_properties)

            if get_from_nullable_chain(job_properties, ["statistics", "numChildJobs"]):
                self.log.debug("Found SCRIPT job. Extracting lineage from child jobs instead.")
                # SCRIPT job type has no input / output information but spawns child jobs that have one
                # https://cloud.google.com/bigquery/docs/information-schema-jobs#multi-statement_query_job
                for child_job_id in self._client.list_jobs(parent_job=self.job_id):
                    child_job_properties = self._client.get_job(job_id=child_job_id)._properties
                    child_inputs, child_outputs = self._get_inputs_and_outputs(child_job_properties)
                    inputs.extend(child_inputs)
                    outputs.extend(child_outputs)
            else:
                inputs, outputs = self._get_inputs_and_outputs(job_properties)

        except Exception as e:
            self.log.warning("Cannot retrieve job details from BigQuery.Client. %s", e, exc_info=True)
            exception_msg = traceback.format_exc()
            run_facets.update(
                {
                    "errorMessage": ErrorMessageRunFacet(
                        message=f"{e}: {exception_msg}",
                        programmingLanguage="python",
                    )
                }
            )

        return OperatorLineage(
            inputs=list(inputs),
            outputs=self._deduplicate_outputs(outputs),
            run_facets=run_facets,
            job_facets={"sql": SQLJobFacet(query=SQLParser.normalize_sql(self.sql))} if self.sql else {},
        )

    def _get_inputs_and_outputs(self, properties: dict) -> tuple[list[InputDataset], list[OutputDataset]]:
        job_type = get_from_nullable_chain(properties, ["configuration", "jobType"])

        if job_type == "QUERY":
            inputs, outputs = self._get_inputs_and_outputs_for_query_job(properties)
        elif job_type == "LOAD":
            inputs, outputs = self._get_inputs_and_outputs_for_load_job(properties)
        elif job_type == "COPY":
            inputs, outputs = self._get_inputs_and_outputs_for_copy_job(properties)
        elif job_type == "EXTRACT":
            inputs, outputs = self._get_inputs_and_outputs_for_extract_job(properties)
        else:
            self.log.debug("Unsupported job type for input/output extraction: `%s`.", job_type)  # type: ignore[attr-defined]
            inputs, outputs = [], []

        return inputs, outputs

    def _deduplicate_outputs(self, outputs: Iterable[OutputDataset | None]) -> list[OutputDataset]:
        final_outputs = {}
        for single_output in outputs:
            if not single_output:
                continue
            key = f"{single_output.namespace}.{single_output.name}"
            if key not in final_outputs:
                final_outputs[key] = single_output
                continue

            # No OutputStatisticsOutputDatasetFacet is added to duplicated outputs as we can not determine
            # if the rowCount or size can be summed together.
            if single_output.outputFacets:
                single_output.outputFacets.pop("outputStatistics", None)

            # If multiple outputs contain Column Level Lineage Facet - merge the facets
            if (
                single_output.facets
                and final_outputs[key].facets
                and "columnLineage" in single_output.facets
                and "columnLineage" in final_outputs[key].facets  # type: ignore
            ):
                single_output.facets["columnLineage"] = merge_column_lineage_facets(
                    [
                        single_output.facets["columnLineage"],  # type: ignore
                        final_outputs[key].facets["columnLineage"],  # type: ignore
                    ]
                )

            final_outputs[key] = single_output

        return list(final_outputs.values())

    def _get_input_dataset(self, table: dict) -> InputDataset:
        return cast("InputDataset", self._get_dataset(table, "input"))

    def _get_output_dataset(self, table: dict) -> OutputDataset:
        return cast("OutputDataset", self._get_dataset(table, "output"))

    def _get_dataset(self, table: dict, dataset_type: str) -> Dataset:
        project = table.get("projectId")
        dataset = table.get("datasetId")
        table_name = table.get("tableId")
        dataset_name = f"{project}.{dataset}.{table_name}"

        dataset_facets = self._get_table_facets_safely(dataset_name)
        if dataset_type == "input":
            # Logic specific to creating InputDataset (if needed)
            return InputDataset(
                namespace=BIGQUERY_NAMESPACE,
                name=dataset_name,
                facets=dataset_facets,
            )
        if dataset_type == "output":
            # Logic specific to creating OutputDataset (if needed)
            return OutputDataset(
                namespace=BIGQUERY_NAMESPACE,
                name=dataset_name,
                facets=dataset_facets,
            )
        raise ValueError("Invalid dataset_type. Must be 'input' or 'output'")

    def _get_table_facets_safely(self, table_name: str) -> dict[str, DatasetFacet]:
        try:
            bq_table = self._client.get_table(table_name)
            return get_facets_from_bq_table(bq_table)
        except Exception as e:
            self.log.warning("Could not extract facets from bigquery table: `%s`. %s", table_name, e)  # type: ignore[attr-defined]
        return {}

    def _get_inputs_and_outputs_for_query_job(
        self, properties: dict
    ) -> tuple[list[InputDataset], list[OutputDataset]]:
        input_tables = get_from_nullable_chain(properties, ["statistics", "query", "referencedTables"]) or []
        output_table = get_from_nullable_chain(properties, ["configuration", "query", "destinationTable"])

        inputs = [
            self._get_input_dataset(input_table)
            for input_table in input_tables
            if input_table != output_table  # Output table is in `referencedTables` and needs to be removed
        ]

        if not output_table:
            return inputs, []

        output = self._get_output_dataset(output_table)
        if dataset_stat_facet := self._get_output_statistics_dataset_facet(properties):
            output.outputFacets = output.outputFacets or {}
            output.outputFacets["outputStatistics"] = dataset_stat_facet
        if cll_facet := self._get_column_level_lineage_facet_for_query_job(properties, output, inputs):
            output.facets = output.facets or {}
            output.facets["columnLineage"] = cll_facet
        return inputs, [output]

    def _get_inputs_and_outputs_for_load_job(
        self, properties: dict
    ) -> tuple[list[InputDataset], list[OutputDataset]]:
        output = self._get_output_dataset(properties["configuration"]["load"]["destinationTable"])
        output_table_schema_facet = output.facets.get("schema") if output.facets else None

        source_uris = properties["configuration"]["load"]["sourceUris"]
        inputs = [
            InputDataset(
                namespace=namespace,
                name=name,
                facets={"schema": output_table_schema_facet} if output_table_schema_facet else {},
            )
            for namespace, name in get_namespace_name_from_source_uris(source_uris)
        ]

        if dataset_stat_facet := self._get_output_statistics_dataset_facet(properties):
            output.outputFacets = output.outputFacets or {}
            output.outputFacets["outputStatistics"] = dataset_stat_facet
        if cll_facet := get_identity_column_lineage_facet(self._extract_column_names(output), inputs):
            output.facets = {**output.facets, **cll_facet} if output.facets else cll_facet
        return inputs, [output]

    def _get_inputs_and_outputs_for_copy_job(
        self, properties: dict
    ) -> tuple[list[InputDataset], list[OutputDataset]]:
        input_tables = get_from_nullable_chain(properties, ["configuration", "copy", "sourceTables"]) or [
            get_from_nullable_chain(properties, ["configuration", "copy", "sourceTable"])
        ]
        inputs = [self._get_input_dataset(input_table) for input_table in input_tables]

        output = self._get_output_dataset(properties["configuration"]["copy"]["destinationTable"])
        if dataset_stat_facet := self._get_output_statistics_dataset_facet(properties):
            output.outputFacets = output.outputFacets or {}
            output.outputFacets["outputStatistics"] = dataset_stat_facet
        if cll_facet := get_identity_column_lineage_facet(self._extract_column_names(output), inputs):
            output.facets = {**output.facets, **cll_facet} if output.facets else cll_facet
        return inputs, [output]

    def _get_inputs_and_outputs_for_extract_job(
        self, properties: dict
    ) -> tuple[list[InputDataset], list[OutputDataset]]:
        source_table = get_from_nullable_chain(properties, ["configuration", "extract", "sourceTable"])
        input_dataset = self._get_input_dataset(source_table) if source_table else None

        destination_uris = get_from_nullable_chain(
            properties, ["configuration", "extract", "destinationUris"]
        ) or [get_from_nullable_chain(properties, ["configuration", "extract", "destinationUri"])]

        outputs = []
        for namespace, name in get_namespace_name_from_source_uris(destination_uris):
            output_facets = {}
            if input_dataset:
                input_schema = input_dataset.facets.get("schema") if input_dataset.facets else None
                if input_schema:
                    output_facets["schema"] = input_schema
                if cll_facet := get_identity_column_lineage_facet(
                    self._extract_column_names(input_dataset), [input_dataset]
                ):
                    output_facets = {**output_facets, **cll_facet}
            outputs.append(OutputDataset(namespace=namespace, name=name, facets=output_facets))

        inputs = [input_dataset] if input_dataset else []
        return inputs, outputs

    @staticmethod
    def _get_bigquery_job_run_facet(properties: dict) -> BigQueryJobRunFacet:
        from airflow.providers.google.cloud.openlineage.facets import BigQueryJobRunFacet

        job_type = get_from_nullable_chain(properties, ["configuration", "jobType"])
        cache_hit, billed_bytes = None, None
        if job_type == "QUERY":
            if get_from_nullable_chain(properties, ["configuration", "query", "query"]):
                # Exclude the query to avoid event size issues and duplicating SqlJobFacet information.
                properties = copy.deepcopy(properties)
                properties["configuration"]["query"].pop("query")
            cache_hit = get_from_nullable_chain(properties, ["statistics", "query", "cacheHit"])
            billed_bytes = get_from_nullable_chain(properties, ["statistics", "query", "totalBytesBilled"])

        return BigQueryJobRunFacet(
            cached=str(cache_hit).lower() == "true",
            billedBytes=int(billed_bytes) if billed_bytes else None,
            properties=json.dumps(properties),
        )

    @staticmethod
    def _get_output_statistics_dataset_facet(
        properties,
    ) -> OutputStatisticsOutputDatasetFacet | None:
        job_type = get_from_nullable_chain(properties, ["configuration", "jobType"])
        out_rows, out_bytes = None, None
        if job_type == "QUERY":
            query_plan = get_from_nullable_chain(properties, chain=["statistics", "query", "queryPlan"])
            if not query_plan:  # Without query plan there is no statistics
                return None
            out_stage = query_plan[-1]  # Last stage of query plan writes the data and has all the statistics
            out_rows = out_stage.get("recordsWritten", None)
            out_bytes = out_stage.get("shuffleOutputBytes", None)
        elif job_type == "LOAD":
            out_rows = get_from_nullable_chain(properties, ["statistics", "load", "outputRows"])
            out_bytes = get_from_nullable_chain(properties, ["statistics", "load", "outputBytes"])
        elif job_type == "COPY":
            out_rows = get_from_nullable_chain(properties, ["statistics", "copy", "copiedRows"])
            out_bytes = get_from_nullable_chain(properties, ["statistics", "copy", "copiedLogicalBytes"])
        # No statistics available for EXTRACT job type

        if out_bytes and out_rows:
            return OutputStatisticsOutputDatasetFacet(rowCount=int(out_rows), size=int(out_bytes))
        return None

    def _get_column_level_lineage_facet_for_query_job(
        self, properties: dict, output: OutputDataset, inputs: Iterable[InputDataset]
    ) -> ColumnLineageDatasetFacet | None:
        """
        Extract column-level lineage information from a BigQuery job and return it as a facet.

        The Column Level Lineage Facet will NOT be returned if any of the following condition is met:
        - The parsed result does not contain column lineage information.
        - The parsed result does not contain exactly one output table.
        - The parsed result has a different output table than the output table from the BQ job.
        - The parsed result has at least one input table not present in the input tables from the BQ job.
        - The parsed result has a column not present in the schema of given dataset from the BQ job.

        Args:
            properties: The properties of the BigQuery job.
            output: The output dataset for which the column lineage is being extracted.

        Returns:
            The extracted Column Lineage Dataset Facet, or None if conditions are not met.
        """
        from airflow.providers.openlineage.sqlparser import SQLParser

        # Extract SQL query and parse it
        self.log.debug("Extracting column-level lineage facet from BigQuery query.")  # type: ignore[attr-defined]

        query = get_from_nullable_chain(properties, ["configuration", "query", "query"])
        if query is None:
            self.log.debug("No query found in BQ job configuration. Facet generation skipped.")  # type: ignore[attr-defined]
            return None

        parse_result = SQLParser("bigquery").parse(SQLParser.split_sql_string(SQLParser.normalize_sql(query)))
        if parse_result is None or parse_result.column_lineage == []:
            self.log.debug("No column-level lineage found in the SQL query. Facet generation skipped.")  # type: ignore[attr-defined]
            return None

        default_dataset, default_project = self._extract_default_dataset_and_project(
            properties,
            self.project_id,  # type: ignore[attr-defined]
        )

        # Verify if the output table id from the parse result matches the BQ job output table
        if not self._validate_output_table_id(
            parse_result,
            output,
            default_project,
            default_dataset,
        ):
            return None

        # Verify if all columns from parse results are present in the output dataset schema
        if not self._validate_output_columns(parse_result, output):
            return None

        input_tables_from_parse_result = self._extract_parsed_input_tables(
            parse_result, default_project, default_dataset
        )
        input_tables_from_bq = {input_ds.name: self._extract_column_names(input_ds) for input_ds in inputs}

        # Verify if all datasets from parse results are present in bq job input datasets
        if not self._validate_input_tables(input_tables_from_parse_result, input_tables_from_bq):
            return None

        # Verify if all columns from parse results are present in their respective bq job input datasets
        if not self._validate_input_columns(input_tables_from_parse_result, input_tables_from_bq):
            return None

        return self._generate_column_lineage_facet(parse_result, default_project, default_dataset)

    @staticmethod
    def _get_qualified_name_from_parse_result(table, default_project: str, default_dataset: str) -> str:
        """Get the qualified name of a table from the parse result."""
        return ".".join(
            (
                table.database or default_project,
                table.schema or default_dataset,
                table.name,
            )
        )

    @staticmethod
    def _extract_default_dataset_and_project(properties: dict, default_project: str) -> tuple[str, str]:
        """Extract the default dataset and project from the BigQuery job properties."""
        default_dataset_obj = get_from_nullable_chain(
            properties, ["configuration", "query", "defaultDataset"]
        )
        default_dataset = default_dataset_obj.get("datasetId", "") if default_dataset_obj else ""
        default_project = (
            default_dataset_obj.get("projectId", default_project) if default_dataset_obj else default_project
        )
        return default_dataset, default_project

    def _validate_output_table_id(
        self, parse_result, output: OutputDataset, default_project: str, default_dataset: str
    ) -> bool:
        """Check if the output table id from the parse result matches the BQ job output table."""
        if len(parse_result.out_tables) != 1:
            self.log.debug(  # type: ignore[attr-defined]
                "Invalid output tables in the parse result: `%s`. Expected exactly one output table.",
                parse_result.out_tables,
            )
            return False

        parsed_output_table = self._get_qualified_name_from_parse_result(
            parse_result.out_tables[0], default_project, default_dataset
        )
        if parsed_output_table != output.name:
            self.log.debug(  # type: ignore[attr-defined]
                "Mismatch between parsed output table `%s` and BQ job output table `%s`.",
                parsed_output_table,
                output.name,
            )
            return False
        return True

    @staticmethod
    def _extract_column_names(dataset: Dataset) -> list[str]:
        """Extract column names from a dataset's schema."""
        return [
            f.name
            for f in dataset.facets.get("schema", SchemaDatasetFacet(fields=[])).fields  # type: ignore[union-attr]
            if dataset.facets
        ]

    def _validate_output_columns(self, parse_result, output: OutputDataset) -> bool:
        """Validate if all descendant columns in parse result exist in output dataset schema."""
        output_column_names = self._extract_column_names(output)
        missing_columns = [
            lineage.descendant.name
            for lineage in parse_result.column_lineage
            if lineage.descendant.name not in output_column_names
        ]
        if missing_columns:
            self.log.debug(  # type: ignore[attr-defined]
                "Output dataset schema is missing columns from the parse result: `%s`.", missing_columns
            )
            return False
        return True

    def _extract_parsed_input_tables(
        self, parse_result, default_project: str, default_dataset: str
    ) -> dict[str, list[str]]:
        """Extract input tables and their columns from the parse result."""
        input_tables: dict[str, list[str]] = {}
        for lineage in parse_result.column_lineage:
            for column_meta in lineage.lineage:
                if not column_meta.origin:
                    self.log.debug(  # type: ignore[attr-defined]
                        "Column `%s` lacks origin information. Skipping facet generation.", column_meta.name
                    )
                    return {}

                input_table_id = self._get_qualified_name_from_parse_result(
                    column_meta.origin, default_project, default_dataset
                )
                input_tables.setdefault(input_table_id, []).append(column_meta.name)
        return input_tables

    def _validate_input_tables(
        self, parsed_input_tables: dict[str, list[str]], input_tables_from_bq: dict[str, list[str]]
    ) -> bool:
        """Validate if all parsed input tables exist in the BQ job's input datasets."""
        if not parsed_input_tables:
            self.log.debug("No input tables found in the parse result. Facet generation skipped.")  # type: ignore[attr-defined]
            return False
        if missing_tables := set(parsed_input_tables) - set(input_tables_from_bq):
            self.log.debug(  # type: ignore[attr-defined]
                "Parsed input tables not found in the BQ job's input datasets: `%s`.", missing_tables
            )
            return False
        return True

    def _validate_input_columns(
        self, parsed_input_tables: dict[str, list[str]], input_tables_from_bq: dict[str, list[str]]
    ) -> bool:
        """Validate if all parsed input columns exist in their respective BQ job input table schemas."""
        if not parsed_input_tables:
            self.log.debug("No input tables found in the parse result. Facet generation skipped.")  # type: ignore[attr-defined]
            return False
        for table, columns in parsed_input_tables.items():
            if missing_columns := set(columns) - set(input_tables_from_bq.get(table, [])):
                self.log.debug(  # type: ignore[attr-defined]
                    "Input table `%s` is missing columns from the parse result: `%s`.", table, missing_columns
                )
                return False
        return True

    def _generate_column_lineage_facet(
        self, parse_result, default_project: str, default_dataset: str
    ) -> ColumnLineageDatasetFacet:
        """Generate the ColumnLineageDatasetFacet based on the parsed result."""
        return ColumnLineageDatasetFacet(
            fields={
                lineage.descendant.name: Fields(
                    inputFields=[
                        InputField(
                            namespace=BIGQUERY_NAMESPACE,
                            name=self._get_qualified_name_from_parse_result(
                                column_meta.origin, default_project, default_dataset
                            ),
                            field=column_meta.name,
                        )
                        for column_meta in lineage.lineage
                    ],
                    transformationType="",
                    transformationDescription="",
                )
                for lineage in parse_result.column_lineage
            }
        )



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

import pytest
from google.cloud import batch_v1

from airflow.exceptions import AirflowException, TaskDeferred
from airflow.providers.google.cloud.operators.cloud_batch import (
    CloudBatchDeleteJobOperator,
    CloudBatchListJobsOperator,
    CloudBatchListTasksOperator,
    CloudBatchSubmitJobOperator,
)

CLOUD_BATCH_HOOK_PATH = "airflow.providers.google.cloud.operators.cloud_batch.CloudBatchHook"
TASK_ID = "test"
PROJECT_ID = "testproject"
REGION = "us-central1"
JOB_NAME = "test"
JOB = batch_v1.Job()
JOB.name = JOB_NAME


class TestCloudBatchSubmitJobOperator:
    @mock.patch(CLOUD_BATCH_HOOK_PATH)
    def test_execute(self, mock):
        mock.return_value.wait_for_job.return_value = JOB
        operator = CloudBatchSubmitJobOperator(
            task_id=TASK_ID, project_id=PROJECT_ID, region=REGION, job_name=JOB_NAME, job=JOB
        )

        completed_job = operator.execute(context=mock.MagicMock())

        assert completed_job["name"] == JOB_NAME

        mock.return_value.submit_batch_job.assert_called_with(
            job_name=JOB_NAME, job=JOB, region=REGION, project_id=PROJECT_ID
        )
        mock.return_value.wait_for_job.assert_called()

    @mock.patch(CLOUD_BATCH_HOOK_PATH)
    def test_execute_deferrable(self, mock):
        operator = CloudBatchSubmitJobOperator(
            task_id=TASK_ID, project_id=PROJECT_ID, region=REGION, job_name=JOB_NAME, job=JOB, deferrable=True
        )

        with pytest.raises(expected_exception=TaskDeferred):
            operator.execute(context=mock.MagicMock())

    @mock.patch(CLOUD_BATCH_HOOK_PATH)
    def test_execute_complete(self, mock):
        mock.return_value.get_job.return_value = JOB
        operator = CloudBatchSubmitJobOperator(
            task_id=TASK_ID, project_id=PROJECT_ID, region=REGION, job_name=JOB_NAME, job=JOB, deferrable=True
        )

        event = {"status": "success", "job_name": JOB_NAME, "message": "test error"}
        completed_job = operator.execute_complete(context=mock.MagicMock(), event=event)

        assert completed_job["name"] == JOB_NAME

        mock.return_value.get_job.assert_called_once_with(job_name=JOB_NAME)

    @mock.patch(CLOUD_BATCH_HOOK_PATH)
    def test_execute_complete_exception(self, mock):
        operator = CloudBatchSubmitJobOperator(
            task_id=TASK_ID, project_id=PROJECT_ID, region=REGION, job_name=JOB_NAME, job=JOB, deferrable=True
        )

        event = {"status": "error", "job_name": JOB_NAME, "message": "test error"}
        with pytest.raises(
            expected_exception=AirflowException, match="Unexpected error in the operation: test error"
        ):
            operator.execute_complete(context=mock.MagicMock(), event=event)


class TestCloudBatchDeleteJobOperator:
    @mock.patch(CLOUD_BATCH_HOOK_PATH)
    def test_execute(self, hook_mock):
        delete_operation_mock = self._delete_operation_mock()
        hook_mock.return_value.delete_job.return_value = delete_operation_mock

        operator = CloudBatchDeleteJobOperator(
            task_id=TASK_ID,
            project_id=PROJECT_ID,
            region=REGION,
            job_name=JOB_NAME,
        )

        operator.execute(context=mock.MagicMock())

        hook_mock.return_value.delete_job.assert_called_once_with(
            job_name=JOB_NAME, region=REGION, project_id=PROJECT_ID
        )
        delete_operation_mock.result.assert_called_once()

    def _delete_operation_mock(self):
        operation = mock.MagicMock()
        operation.result.return_value = mock.MagicMock()
        return operation


class TestCloudBatchListJobsOperator:
    @mock.patch(CLOUD_BATCH_HOOK_PATH)
    def test_execute(self, hook_mock):
        filter = "filter_description"
        limit = 2
        operator = CloudBatchListJobsOperator(
            task_id=TASK_ID, project_id=PROJECT_ID, region=REGION, filter=filter, limit=limit
        )

        operator.execute(context=mock.MagicMock())

        hook_mock.return_value.list_jobs.assert_called_once_with(
            region=REGION, project_id=PROJECT_ID, filter=filter, limit=limit
        )

    @mock.patch(CLOUD_BATCH_HOOK_PATH)
    def test_execute_with_invalid_limit(self, hook_mock):
        filter = "filter_description"
        limit = -1

        with pytest.raises(expected_exception=AirflowException):
            CloudBatchListJobsOperator(
                task_id=TASK_ID, project_id=PROJECT_ID, region=REGION, filter=filter, limit=limit
            )


class TestCloudBatchListTasksOperator:
    @mock.patch(CLOUD_BATCH_HOOK_PATH)
    def test_execute(self, hook_mock):
        filter = "filter_description"
        limit = 2
        job_name = "test_job"

        operator = CloudBatchListTasksOperator(
            task_id=TASK_ID,
            project_id=PROJECT_ID,
            region=REGION,
            job_name=job_name,
            filter=filter,
            limit=limit,
        )

        operator.execute(context=mock.MagicMock())

        hook_mock.return_value.list_tasks.assert_called_once_with(
            region=REGION,
            project_id=PROJECT_ID,
            filter=filter,
            job_name=job_name,
            limit=limit,
            group_name="group0",
        )

    @mock.patch(CLOUD_BATCH_HOOK_PATH)
    def test_execute_with_invalid_limit(self, hook_mock):
        filter = "filter_description"
        limit = -1
        job_name = "test_job"

        with pytest.raises(expected_exception=AirflowException):
            CloudBatchListTasksOperator(
                task_id=TASK_ID,
                project_id=PROJECT_ID,
                region=REGION,
                job_name=job_name,
                filter=filter,
                limit=limit,
            )



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING

from google.cloud.batch_v1 import Job, Task

from airflow.configuration import conf
from airflow.exceptions import AirflowException
from airflow.providers.google.cloud.hooks.cloud_batch import CloudBatchHook
from airflow.providers.google.cloud.operators.cloud_base import GoogleCloudBaseOperator
from airflow.providers.google.cloud.triggers.cloud_batch import CloudBatchJobFinishedTrigger

if TYPE_CHECKING:
    from google.api_core import operation

    from airflow.providers.common.compat.sdk import Context


class CloudBatchSubmitJobOperator(GoogleCloudBaseOperator):
    """
    Submit a job and wait for its completion.

    :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
    :param region: Required. The ID of the Google Cloud region that the service belongs to.
    :param job_name: Required. The name of the job to create.
    :param job: Required. The job descriptor containing the configuration of the job to submit.
    :param polling_period_seconds: Optional: Control the rate of the poll for the result of deferrable run.
        By default, the trigger will poll every 10 seconds.
    :param timeout: The timeout for this request.
    :param gcp_conn_id: The connection ID used to connect to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    :param deferrable: Run operator in the deferrable mode

    """

    template_fields = ("project_id", "region", "gcp_conn_id", "impersonation_chain", "job_name")

    def __init__(
        self,
        project_id: str,
        region: str,
        job_name: str,
        job: dict | Job,
        polling_period_seconds: float = 10,
        timeout_seconds: float | None = None,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        deferrable: bool = conf.getboolean("operators", "default_deferrable", fallback=False),
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.project_id = project_id
        self.region = region
        self.job_name = job_name
        self.job = job
        self.polling_period_seconds = polling_period_seconds
        self.timeout_seconds = timeout_seconds
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain
        self.deferrable = deferrable
        self.polling_period_seconds = polling_period_seconds

    def execute(self, context: Context):
        hook: CloudBatchHook = CloudBatchHook(self.gcp_conn_id, self.impersonation_chain)
        job = hook.submit_batch_job(
            job_name=self.job_name, job=self.job, region=self.region, project_id=self.project_id
        )

        if not self.deferrable:
            completed_job = hook.wait_for_job(
                job_name=job.name,
                polling_period_seconds=self.polling_period_seconds,
                timeout=self.timeout_seconds,
            )

            return Job.to_dict(completed_job)

        self.defer(
            trigger=CloudBatchJobFinishedTrigger(
                job_name=job.name,
                project_id=self.project_id,
                gcp_conn_id=self.gcp_conn_id,
                impersonation_chain=self.impersonation_chain,
                location=self.region,
                polling_period_seconds=self.polling_period_seconds,
                timeout=self.timeout_seconds,
            ),
            method_name="execute_complete",
        )

    def execute_complete(self, context: Context, event: dict):
        job_status = event["status"]
        if job_status == "success":
            hook: CloudBatchHook = CloudBatchHook(self.gcp_conn_id, self.impersonation_chain)
            job = hook.get_job(job_name=event["job_name"])
            return Job.to_dict(job)
        raise AirflowException(f"Unexpected error in the operation: {event['message']}")


class CloudBatchDeleteJobOperator(GoogleCloudBaseOperator):
    """
    Deletes a job and wait for the operation to be completed.

    :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
    :param region: Required. The ID of the Google Cloud region that the service belongs to.
    :param job_name: Required. The name of the job to be deleted.
    :param timeout: The timeout for this request.
    :param gcp_conn_id: The connection ID used to connect to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).

    """

    template_fields = ("project_id", "region", "gcp_conn_id", "impersonation_chain", "job_name")

    def __init__(
        self,
        project_id: str,
        region: str,
        job_name: str,
        timeout: float | None = None,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.project_id = project_id
        self.region = region
        self.job_name = job_name
        self.timeout = timeout
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

    def execute(self, context: Context):
        hook: CloudBatchHook = CloudBatchHook(self.gcp_conn_id, self.impersonation_chain)

        operation = hook.delete_job(job_name=self.job_name, region=self.region, project_id=self.project_id)

        self._wait_for_operation(operation)

    def _wait_for_operation(self, operation: operation.Operation):
        try:
            return operation.result(timeout=self.timeout)
        except Exception:
            error = operation.exception(timeout=self.timeout)
            raise AirflowException(error)


class CloudBatchListJobsOperator(GoogleCloudBaseOperator):
    """
    List Cloud Batch jobs.

    :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
    :param region: Required. The ID of the Google Cloud region that the service belongs to.
    :param gcp_conn_id: The connection ID used to connect to Google Cloud.
    :param filter: The filter based on which to list the jobs. If left empty, all the jobs are listed.
    :param limit: The number of jobs to list. If left empty,
        all the jobs matching the filter will be returned.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).

    """

    template_fields = (
        "project_id",
        "region",
        "gcp_conn_id",
        "impersonation_chain",
    )

    def __init__(
        self,
        project_id: str,
        region: str,
        gcp_conn_id: str = "google_cloud_default",
        filter: str | None = None,
        limit: int | None = None,
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.project_id = project_id
        self.region = region
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain
        self.filter = filter
        self.limit = limit
        if limit is not None and limit < 0:
            raise AirflowException("The limit for the list jobs request should be greater or equal to zero")

    def execute(self, context: Context):
        hook: CloudBatchHook = CloudBatchHook(self.gcp_conn_id, self.impersonation_chain)

        jobs_list = hook.list_jobs(
            region=self.region, project_id=self.project_id, filter=self.filter, limit=self.limit
        )

        return [Job.to_dict(job) for job in jobs_list]


class CloudBatchListTasksOperator(GoogleCloudBaseOperator):
    """
    List Cloud Batch tasks for a given job.

    :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
    :param region: Required. The ID of the Google Cloud region that the service belongs to.
    :param job_name: Required. The name of the job for which to list tasks.
    :param gcp_conn_id: The connection ID used to connect to Google Cloud.
    :param filter: The filter based on which to list the jobs. If left empty, all the jobs are listed.
    :param group_name: The name of the group that owns the task. By default, it's `group0`.
    :param limit: The number of tasks to list.
        If left empty, all the tasks matching the filter will be returned.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).

    """

    template_fields = ("project_id", "region", "job_name", "gcp_conn_id", "impersonation_chain", "group_name")

    def __init__(
        self,
        project_id: str,
        region: str,
        job_name: str,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        group_name: str = "group0",
        filter: str | None = None,
        limit: int | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.project_id = project_id
        self.region = region
        self.job_name = job_name
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain
        self.group_name = group_name
        self.filter = filter
        self.limit = limit
        if limit is not None and limit < 0:
            raise AirflowException("The limit for the list jobs request should be greater or equal to zero")

    def execute(self, context: Context):
        hook: CloudBatchHook = CloudBatchHook(self.gcp_conn_id, self.impersonation_chain)

        tasks_list = hook.list_tasks(
            region=self.region,
            project_id=self.project_id,
            job_name=self.job_name,
            group_name=self.group_name,
            filter=self.filter,
            limit=self.limit,
        )

        return [Task.to_dict(task) for task in tasks_list]



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

import pytest
from google.cloud.run_v2 import (
    CreateJobRequest,
    CreateServiceRequest,
    DeleteJobRequest,
    DeleteServiceRequest,
    GetJobRequest,
    GetServiceRequest,
    Job,
    ListJobsRequest,
    RunJobRequest,
    Service,
    UpdateJobRequest,
)
from google.longrunning import operations_pb2

from airflow.exceptions import AirflowException
from airflow.providers.google.cloud.hooks.cloud_run import (
    CloudRunAsyncHook,
    CloudRunHook,
    CloudRunServiceAsyncHook,
    CloudRunServiceHook,
)

from unit.google.cloud.utils.base_gcp_mock import mock_base_gcp_hook_default_project_id

PROJECT_ID = "projectid"
REGION = "region1"
JOB_NAME = "job1"
SERVICE_NAME = "service1"
OPERATION_NAME = "operationname"


@pytest.mark.db_test
class TestCloudRunHook:
    def dummy_get_credentials(self):
        pass

    @pytest.fixture
    def cloud_run_hook(self):
        cloud_run_hook = CloudRunHook()
        cloud_run_hook.get_credentials = self.dummy_get_credentials
        return cloud_run_hook

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.__init__",
        new=mock_base_gcp_hook_default_project_id,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_run.JobsClient")
    def test_get_job(self, mock_batch_service_client, cloud_run_hook):
        get_job_request = GetJobRequest(name=f"projects/{PROJECT_ID}/locations/{REGION}/jobs/{JOB_NAME}")

        cloud_run_hook.get_job(job_name=JOB_NAME, region=REGION, project_id=PROJECT_ID)
        cloud_run_hook._client.get_job.assert_called_once_with(get_job_request)

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.__init__",
        new=mock_base_gcp_hook_default_project_id,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_run.JobsClient")
    def test_update_job(self, mock_batch_service_client, cloud_run_hook):
        job = Job()
        job.name = f"projects/{PROJECT_ID}/locations/{REGION}/jobs/{JOB_NAME}"

        update_request = UpdateJobRequest()
        update_request.job = job

        cloud_run_hook.update_job(
            job=Job.to_dict(job), job_name=JOB_NAME, region=REGION, project_id=PROJECT_ID
        )

        cloud_run_hook._client.update_job.assert_called_once_with(update_request)

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.__init__",
        new=mock_base_gcp_hook_default_project_id,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_run.JobsClient")
    def test_create_job(self, mock_batch_service_client, cloud_run_hook):
        job = Job()

        create_request = CreateJobRequest()
        create_request.job = job
        create_request.job_id = JOB_NAME
        create_request.parent = f"projects/{PROJECT_ID}/locations/{REGION}"

        cloud_run_hook.create_job(
            job=Job.to_dict(job), job_name=JOB_NAME, region=REGION, project_id=PROJECT_ID
        )

        cloud_run_hook._client.create_job.assert_called_once_with(create_request)

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.__init__",
        new=mock_base_gcp_hook_default_project_id,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_run.JobsClient")
    def test_execute_job(self, mock_batch_service_client, cloud_run_hook):
        overrides = {
            "container_overrides": [{"args": ["python", "main.py"]}],
            "task_count": 1,
            "timeout": "60s",
        }
        run_job_request = RunJobRequest(
            name=f"projects/{PROJECT_ID}/locations/{REGION}/jobs/{JOB_NAME}", overrides=overrides
        )

        cloud_run_hook.execute_job(
            job_name=JOB_NAME, region=REGION, project_id=PROJECT_ID, overrides=overrides
        )
        cloud_run_hook._client.run_job.assert_called_once_with(request=run_job_request)

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.__init__",
        new=mock_base_gcp_hook_default_project_id,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_run.JobsClient")
    def test_list_jobs(self, mock_batch_service_client, cloud_run_hook):
        number_of_jobs = 3
        region = "us-central1"
        project_id = "test_project_id"

        page = self._mock_pager(number_of_jobs)
        mock_batch_service_client.return_value.list_jobs.return_value = page

        jobs_list = cloud_run_hook.list_jobs(region=region, project_id=project_id)

        for i in range(number_of_jobs):
            assert jobs_list[i].name == f"name{i}"

        expected_list_jobs_request: ListJobsRequest = ListJobsRequest(
            parent=f"projects/{project_id}/locations/{region}"
        )
        mock_batch_service_client.return_value.list_jobs.assert_called_once_with(
            request=expected_list_jobs_request
        )

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.__init__",
        new=mock_base_gcp_hook_default_project_id,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_run.JobsClient")
    def test_list_jobs_show_deleted(self, mock_batch_service_client, cloud_run_hook):
        number_of_jobs = 3
        region = "us-central1"
        project_id = "test_project_id"

        page = self._mock_pager(number_of_jobs)
        mock_batch_service_client.return_value.list_jobs.return_value = page

        jobs_list = cloud_run_hook.list_jobs(region=region, project_id=project_id, show_deleted=True)

        for i in range(number_of_jobs):
            assert jobs_list[i].name == f"name{i}"

        expected_list_jobs_request: ListJobsRequest = ListJobsRequest(
            parent=f"projects/{project_id}/locations/{region}", show_deleted=True
        )
        mock_batch_service_client.return_value.list_jobs.assert_called_once_with(
            request=expected_list_jobs_request
        )

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.__init__",
        new=mock_base_gcp_hook_default_project_id,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_run.JobsClient")
    def test_list_jobs_with_limit(self, mock_batch_service_client, cloud_run_hook):
        number_of_jobs = 3
        limit = 2
        region = "us-central1"
        project_id = "test_project_id"

        page = self._mock_pager(number_of_jobs)
        mock_batch_service_client.return_value.list_jobs.return_value = page

        jobs_list = cloud_run_hook.list_jobs(region=region, project_id=project_id, limit=limit)

        assert len(jobs_list) == limit
        for i in range(limit):
            assert jobs_list[i].name == f"name{i}"

    @mock.patch("airflow.providers.google.cloud.hooks.cloud_run.JobsClient")
    def test_list_jobs_with_limit_zero(self, mock_batch_service_client, cloud_run_hook):
        number_of_jobs = 3
        limit = 0
        region = "us-central1"
        project_id = "test_project_id"

        page = self._mock_pager(number_of_jobs)
        mock_batch_service_client.return_value.list_jobs.return_value = page

        jobs_list = cloud_run_hook.list_jobs(region=region, project_id=project_id, limit=limit)

        assert len(jobs_list) == 0

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.__init__",
        new=mock_base_gcp_hook_default_project_id,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_run.JobsClient")
    def test_list_jobs_with_limit_greater_then_range(self, mock_batch_service_client, cloud_run_hook):
        number_of_jobs = 3
        limit = 5
        region = "us-central1"
        project_id = "test_project_id"

        page = self._mock_pager(number_of_jobs)
        mock_batch_service_client.return_value.list_jobs.return_value = page

        jobs_list = cloud_run_hook.list_jobs(region=region, project_id=project_id, limit=limit)

        assert len(jobs_list) == number_of_jobs
        for i in range(number_of_jobs):
            assert jobs_list[i].name == f"name{i}"

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseHook.__init__",
        new=mock_base_gcp_hook_default_project_id,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_run.JobsClient")
    def test_list_jobs_with_limit_less_than_zero(self, mock_batch_service_client, cloud_run_hook):
        number_of_jobs = 3
        limit = -1
        region = "us-central1"
        project_id = "test_project_id"

        page = self._mock_pager(number_of_jobs)
        mock_batch_service_client.return_value.list_jobs.return_value = page

        with pytest.raises(expected_exception=AirflowException):
            cloud_run_hook.list_jobs(region=region, project_id=project_id, limit=limit)

    @mock.patch("airflow.providers.google.cloud.hooks.cloud_run.JobsClient")
    def test_delete_job(self, mock_batch_service_client, cloud_run_hook):
        delete_request = DeleteJobRequest(name=f"projects/{PROJECT_ID}/locations/{REGION}/jobs/{JOB_NAME}")

        cloud_run_hook.delete_job(job_name=JOB_NAME, region=REGION, project_id=PROJECT_ID)
        cloud_run_hook._client.delete_job.assert_called_once_with(delete_request)

    def _mock_pager(self, number_of_jobs):
        mock_pager = []
        for i in range(number_of_jobs):
            mock_pager.append(Job(name=f"name{i}"))

        return mock_pager


class TestCloudRunAsyncHook:
    @pytest.mark.asyncio
    async def test_get_operation(self):
        hook = CloudRunAsyncHook()
        hook.get_conn = mock.AsyncMock()
        await hook.get_operation(operation_name=OPERATION_NAME)
        hook.get_conn.return_value.get_operation.assert_called_once_with(
            operations_pb2.GetOperationRequest(name=OPERATION_NAME), timeout=120
        )

    def mock_get_operation(self, expected_operation):
        get_operation_mock = mock.AsyncMock()
        get_operation_mock.return_value = expected_operation
        return get_operation_mock

    def _dummy_get_credentials(self):
        pass


@pytest.mark.db_test
class TestCloudRunServiceHook:
    def dummy_get_credentials(self):
        pass

    @pytest.fixture
    def cloud_run_service_hook(self):
        cloud_run_service_hook = CloudRunServiceHook()
        cloud_run_service_hook.get_credentials = self.dummy_get_credentials
        return cloud_run_service_hook

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseAsyncHook.__init__",
        new=mock_base_gcp_hook_default_project_id,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_run.ServicesClient")
    def test_get_service(self, mock_batch_service_client, cloud_run_service_hook):
        get_service_request = GetServiceRequest(
            name=f"projects/{PROJECT_ID}/locations/{REGION}/services/{SERVICE_NAME}"
        )

        cloud_run_service_hook.get_service(service_name=SERVICE_NAME, region=REGION, project_id=PROJECT_ID)
        cloud_run_service_hook._client.get_service.assert_called_once_with(get_service_request)

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseAsyncHook.__init__",
        new=mock_base_gcp_hook_default_project_id,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_run.ServicesClient")
    def test_create_service(self, mock_batch_service_client, cloud_run_service_hook):
        service = Service()

        create_request = CreateServiceRequest()
        create_request.service = service
        create_request.service_id = SERVICE_NAME
        create_request.parent = f"projects/{PROJECT_ID}/locations/{REGION}"

        cloud_run_service_hook.create_service(
            service=service, service_name=SERVICE_NAME, region=REGION, project_id=PROJECT_ID
        )
        cloud_run_service_hook._client.create_service.assert_called_once_with(create_request)

    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseAsyncHook.__init__",
        new=mock_base_gcp_hook_default_project_id,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_run.ServicesClient")
    def test_delete_service(self, mock_batch_service_client, cloud_run_service_hook):
        delete_request = DeleteServiceRequest(
            name=f"projects/{PROJECT_ID}/locations/{REGION}/services/{SERVICE_NAME}"
        )

        cloud_run_service_hook.delete_service(service_name=SERVICE_NAME, region=REGION, project_id=PROJECT_ID)
        cloud_run_service_hook._client.delete_service.assert_called_once_with(delete_request)


class TestCloudRunServiceAsyncHook:
    def dummy_get_credentials(self):
        pass

    def mock_service(self):
        return mock.AsyncMock()

    @pytest.mark.asyncio
    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseAsyncHook.__init__",
        new=mock_base_gcp_hook_default_project_id,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_run.ServicesAsyncClient")
    async def test_create_service(self, mock_client):
        mock_client.return_value = mock.MagicMock()
        mock_client.return_value.create_service = self.mock_service()

        hook = CloudRunServiceAsyncHook()
        hook.get_credentials = self.dummy_get_credentials

        await hook.create_service(
            service_name=SERVICE_NAME,
            service=Service(),
            region=REGION,
            project_id=PROJECT_ID,
        )

        expected_request = CreateServiceRequest(
            service=Service(),
            service_id=SERVICE_NAME,
            parent=f"projects/{PROJECT_ID}/locations/{REGION}",
        )

        mock_client.return_value.create_service.assert_called_once_with(expected_request)

    @pytest.mark.asyncio
    @mock.patch(
        "airflow.providers.google.common.hooks.base_google.GoogleBaseAsyncHook.__init__",
        new=mock_base_gcp_hook_default_project_id,
    )
    @mock.patch("airflow.providers.google.cloud.hooks.cloud_run.ServicesAsyncClient")
    async def test_delete_service(self, mock_client):
        mock_client.return_value = mock.MagicMock()
        mock_client.return_value.delete_service = self.mock_service()

        hook = CloudRunServiceAsyncHook()
        hook.get_credentials = self.dummy_get_credentials

        await hook.delete_service(
            service_name=SERVICE_NAME,
            region=REGION,
            project_id=PROJECT_ID,
        )

        expected_request = DeleteServiceRequest(
            name=f"projects/{PROJECT_ID}/locations/{REGION}/services/{SERVICE_NAME}",
        )

        mock_client.return_value.delete_service.assert_called_once_with(expected_request)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock
from unittest.mock import PropertyMock

import pytest
import time_machine

from airflow.exceptions import AirflowException, TaskDeferred
from airflow.providers.google.cloud.hooks.cloud_storage_transfer_service import CloudDataTransferServiceHook
from airflow.providers.google.cloud.transfers.s3_to_gcs import S3ToGCSOperator
from airflow.utils.timezone import utcnow

PROJECT_ID = "test-project-id"
TASK_ID = "test-s3-gcs-operator"
S3_BUCKET = "test-bucket"
S3_PREFIX = "TEST"
S3_DELIMITER = "/"
GCS_BUCKET = "gcs-bucket"
GCS_BUCKET_URI = "gs://" + GCS_BUCKET
GCS_PREFIX = "data/"
GCS_BUCKET = "gcs-bucket"
GCS_BLOB = "data/"
GCS_PATH_PREFIX = f"gs://{GCS_BUCKET}/{GCS_BLOB}"
MOCK_FILE_1 = "TEST1.csv"
MOCK_FILE_2 = "TEST2.csv"
MOCK_FILE_3 = "TEST3.csv"
MOCK_FILES = [MOCK_FILE_1, MOCK_FILE_2, MOCK_FILE_3]
AWS_CONN_ID = "aws_default"
AWS_ACCESS_KEY_ID = "Mock AWS access key id"
AWS_SECRET_ACCESS_KEY = "Mock AWS secret access key"
GCS_CONN_ID = "google_cloud_default"
IMPERSONATION_CHAIN = ["ACCOUNT_1", "ACCOUNT_2", "ACCOUNT_3"]
DEFERRABLE = False
POLL_INTERVAL = 10
TRANSFER_JOB_ID_0 = "test-transfer-job-0"
TRANSFER_JOB_ID_1 = "test-transfer-job-1"
TRANSFER_JOBS = [TRANSFER_JOB_ID_0, TRANSFER_JOB_ID_1]
APPLY_GCS_PREFIX = False
PARAMETRIZED_OBJECT_PATHS = (
    "apply_gcs_prefix, s3_prefix, s3_object, gcs_destination, gcs_object",
    [
        (False, "", MOCK_FILE_1, GCS_PATH_PREFIX, GCS_PREFIX + MOCK_FILE_1),
        (False, S3_PREFIX, MOCK_FILE_1, GCS_PATH_PREFIX, GCS_PREFIX + S3_PREFIX + MOCK_FILE_1),
        (False, "", MOCK_FILE_1, GCS_BUCKET_URI, MOCK_FILE_1),
        (False, S3_PREFIX, MOCK_FILE_1, GCS_BUCKET_URI, S3_PREFIX + MOCK_FILE_1),
        (True, "", MOCK_FILE_1, GCS_PATH_PREFIX, GCS_PREFIX + MOCK_FILE_1),
        (True, S3_PREFIX, MOCK_FILE_1, GCS_PATH_PREFIX, GCS_PREFIX + MOCK_FILE_1),
        (True, "", MOCK_FILE_1, GCS_BUCKET_URI, MOCK_FILE_1),
        (True, S3_PREFIX, MOCK_FILE_1, GCS_BUCKET_URI, MOCK_FILE_1),
    ],
)


class TestS3ToGoogleCloudStorageOperator:
    def test_init(self):
        """Test S3ToGCSOperator instance is properly initialized."""

        operator = S3ToGCSOperator(
            task_id=TASK_ID,
            bucket=S3_BUCKET,
            prefix=S3_PREFIX,
            delimiter=S3_DELIMITER,
            gcp_conn_id=GCS_CONN_ID,
            dest_gcs=GCS_PATH_PREFIX,
            google_impersonation_chain=IMPERSONATION_CHAIN,
            apply_gcs_prefix=APPLY_GCS_PREFIX,
            deferrable=DEFERRABLE,
            poll_interval=POLL_INTERVAL,
        )

        assert operator.task_id == TASK_ID
        assert operator.bucket == S3_BUCKET
        assert operator.prefix == S3_PREFIX
        assert operator.delimiter == S3_DELIMITER
        assert operator.gcp_conn_id == GCS_CONN_ID
        assert operator.dest_gcs == GCS_PATH_PREFIX
        assert operator.google_impersonation_chain == IMPERSONATION_CHAIN
        assert operator.apply_gcs_prefix == APPLY_GCS_PREFIX
        assert operator.deferrable == DEFERRABLE
        assert operator.poll_interval == POLL_INTERVAL

    @mock.patch("airflow.providers.google.cloud.transfers.s3_to_gcs.S3Hook")
    @mock.patch("airflow.providers.google.cloud.transfers.s3_to_gcs.GCSHook")
    def test_execute(self, gcs_mock_hook, s3_mock_hook):
        """Test the execute function when the run is successful."""

        operator = S3ToGCSOperator(
            task_id=TASK_ID,
            bucket=S3_BUCKET,
            prefix=S3_PREFIX,
            delimiter=S3_DELIMITER,
            gcp_conn_id=GCS_CONN_ID,
            dest_gcs=GCS_PATH_PREFIX,
            google_impersonation_chain=IMPERSONATION_CHAIN,
        )
        operator.hook = mock.MagicMock()

        operator.hook.list_keys.return_value = MOCK_FILES

        uploaded_files = operator.execute(context={})
        gcs_mock_hook.return_value.upload.assert_has_calls(
            [
                mock.call(GCS_BUCKET, GCS_PREFIX + MOCK_FILE_1, mock.ANY, gzip=False),
                mock.call(GCS_BUCKET, GCS_PREFIX + MOCK_FILE_2, mock.ANY, gzip=False),
                mock.call(GCS_BUCKET, GCS_PREFIX + MOCK_FILE_3, mock.ANY, gzip=False),
            ],
            any_order=True,
        )

        operator.hook.list_keys.assert_called_once()
        s3_mock_hook.assert_called_once_with(aws_conn_id=AWS_CONN_ID, verify=None)
        gcs_mock_hook.assert_called_once_with(
            gcp_conn_id=GCS_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )

        # we expect MOCK_FILES to be uploaded
        assert sorted(MOCK_FILES) == sorted(uploaded_files)

    @mock.patch("airflow.providers.google.cloud.transfers.s3_to_gcs.S3Hook")
    @mock.patch("airflow.providers.google.cloud.transfers.s3_to_gcs.GCSHook")
    def test_execute_with_gzip(self, gcs_mock_hook, s3_mock_hook):
        """Test the execute function when the run is successful."""

        operator = S3ToGCSOperator(
            task_id=TASK_ID,
            bucket=S3_BUCKET,
            prefix=S3_PREFIX,
            delimiter=S3_DELIMITER,
            gcp_conn_id=GCS_CONN_ID,
            dest_gcs=GCS_PATH_PREFIX,
            gzip=True,
        )

        operator.hook = mock.MagicMock()

        operator.hook.list_keys.return_value = MOCK_FILES

        operator.execute(context={})
        gcs_mock_hook.assert_called_once_with(
            gcp_conn_id=GCS_CONN_ID,
            impersonation_chain=None,
        )
        gcs_mock_hook.return_value.upload.assert_has_calls(
            [
                mock.call(GCS_BUCKET, GCS_PREFIX + MOCK_FILE_1, mock.ANY, gzip=True),
                mock.call(GCS_BUCKET, GCS_PREFIX + MOCK_FILE_2, mock.ANY, gzip=True),
                mock.call(GCS_BUCKET, GCS_PREFIX + MOCK_FILE_3, mock.ANY, gzip=True),
            ],
            any_order=True,
        )

    @pytest.mark.parametrize(
        ("source_objects", "existing_objects", "objects_expected"),
        [
            (MOCK_FILES, [], MOCK_FILES),
            (MOCK_FILES, [MOCK_FILE_1], [MOCK_FILE_2, MOCK_FILE_3]),
            (MOCK_FILES, [MOCK_FILE_1, MOCK_FILE_2], [MOCK_FILE_3]),
            (MOCK_FILES, [MOCK_FILE_3, MOCK_FILE_2], [MOCK_FILE_1]),
            (MOCK_FILES, MOCK_FILES, []),
        ],
    )
    @mock.patch("airflow.providers.google.cloud.transfers.s3_to_gcs.GCSHook")
    def test_exclude_existing_objects(
        self, mock_gcs_hook, source_objects, existing_objects, objects_expected
    ):
        operator = S3ToGCSOperator(
            task_id=TASK_ID,
            bucket=S3_BUCKET,
            prefix=S3_PREFIX,
            delimiter=S3_DELIMITER,
            gcp_conn_id=GCS_CONN_ID,
            dest_gcs=GCS_PATH_PREFIX,
            gzip=True,
        )
        mock_gcs_hook.list.return_value = existing_objects
        files_reduced = operator.exclude_existing_objects(s3_objects=source_objects, gcs_hook=mock_gcs_hook)
        assert set(files_reduced) == set(objects_expected)

    @pytest.mark.parametrize(*PARAMETRIZED_OBJECT_PATHS)
    def test_s3_to_gcs_object(self, apply_gcs_prefix, s3_prefix, s3_object, gcs_destination, gcs_object):
        operator = S3ToGCSOperator(
            task_id=TASK_ID,
            bucket=S3_BUCKET,
            prefix=s3_prefix,
            delimiter=S3_DELIMITER,
            gcp_conn_id=GCS_CONN_ID,
            dest_gcs=gcs_destination,
            gzip=True,
            apply_gcs_prefix=apply_gcs_prefix,
        )
        assert operator.s3_to_gcs_object(s3_object=s3_prefix + s3_object) == gcs_object

    @pytest.mark.parametrize(*PARAMETRIZED_OBJECT_PATHS)
    def test_gcs_to_s3_object(self, apply_gcs_prefix, s3_prefix, s3_object, gcs_destination, gcs_object):
        operator = S3ToGCSOperator(
            task_id=TASK_ID,
            bucket=S3_BUCKET,
            prefix=s3_prefix,
            delimiter=S3_DELIMITER,
            gcp_conn_id=GCS_CONN_ID,
            dest_gcs=gcs_destination,
            gzip=True,
            apply_gcs_prefix=apply_gcs_prefix,
        )
        assert operator.gcs_to_s3_object(gcs_object=gcs_object) == s3_prefix + s3_object

    @pytest.mark.parametrize(*PARAMETRIZED_OBJECT_PATHS)
    @mock.patch("airflow.providers.google.cloud.transfers.s3_to_gcs.S3Hook")
    @mock.patch("airflow.providers.google.cloud.transfers.s3_to_gcs.GCSHook")
    def test_execute_apply_gcs_prefix(
        self,
        gcs_mock_hook,
        s3_mock_hook,
        apply_gcs_prefix,
        s3_prefix,
        s3_object,
        gcs_destination,
        gcs_object,
    ):
        operator = S3ToGCSOperator(
            task_id=TASK_ID,
            bucket=S3_BUCKET,
            prefix=s3_prefix,
            delimiter=S3_DELIMITER,
            gcp_conn_id=GCS_CONN_ID,
            dest_gcs=gcs_destination,
            google_impersonation_chain=IMPERSONATION_CHAIN,
            apply_gcs_prefix=apply_gcs_prefix,
        )
        operator.hook = mock.MagicMock()
        operator.hook.list_keys.return_value = [s3_prefix + s3_object]

        uploaded_files = operator.execute(context={})
        gcs_mock_hook.return_value.upload.assert_has_calls(
            [
                mock.call(GCS_BUCKET, gcs_object, mock.ANY, gzip=False),
            ],
            any_order=True,
        )

        operator.hook.list_keys.assert_called_once()
        s3_mock_hook.assert_called_once_with(aws_conn_id=AWS_CONN_ID, verify=None)
        gcs_mock_hook.assert_called_once_with(
            gcp_conn_id=GCS_CONN_ID,
            impersonation_chain=IMPERSONATION_CHAIN,
        )

        assert sorted([s3_prefix + s3_object]) == sorted(uploaded_files)

    @pytest.mark.parametrize(
        ("s3_prefix", "gcs_destination", "apply_gcs_prefix", "expected_input", "expected_output"),
        [
            ("dir/pre", "gs://bucket/dest_dir/", False, "dir/pre", "dest_dir/dir"),
            ("dir/pre/", "gs://bucket/dest_dir/", False, "dir/pre", "dest_dir/dir/pre"),
            ("dir/pre", "gs://bucket/dest_dir/", True, "dir/pre", "dest_dir"),
            ("dir/pre", "gs://bucket/", False, "dir/pre", "dir"),
            ("dir/pre", "gs://bucket/", True, "dir/pre", "/"),
            ("", "gs://bucket/", False, "/", "/"),
            ("", "gs://bucket/", True, "/", "/"),
        ],
    )
    def test_get_openlineage_facets_on_start(
        self, s3_prefix, gcs_destination, apply_gcs_prefix, expected_input, expected_output
    ):
        operator = S3ToGCSOperator(
            task_id=TASK_ID,
            bucket=S3_BUCKET,
            prefix=s3_prefix,
            dest_gcs=gcs_destination,
            apply_gcs_prefix=apply_gcs_prefix,
        )
        result = operator.get_openlineage_facets_on_start()
        assert not result.job_facets
        assert not result.run_facets
        assert len(result.outputs) == 1
        assert len(result.inputs) == 1
        assert result.outputs[0].namespace == "gs://bucket"
        assert result.outputs[0].name == expected_output
        assert result.inputs[0].namespace == f"s3://{S3_BUCKET}"
        assert result.inputs[0].name == expected_input


class TestS3ToGoogleCloudStorageOperatorDeferrable:
    @mock.patch("airflow.providers.google.cloud.transfers.s3_to_gcs.CloudDataTransferServiceHook")
    @mock.patch("airflow.providers.google.cloud.transfers.s3_to_gcs.S3Hook")
    @mock.patch("airflow.providers.google.cloud.transfers.s3_to_gcs.GCSHook")
    def test_execute_deferrable(self, mock_gcs_hook, mock_s3_hook, mock_transfer_hook):
        mock_gcs_hook.return_value.project_id = PROJECT_ID

        mock_s3_super_hook = mock.MagicMock()
        mock_s3_super_hook.list_keys.return_value = MOCK_FILES
        mock_s3_hook.conn_config = mock.MagicMock(
            aws_access_key_id=AWS_ACCESS_KEY_ID,
            aws_secret_access_key=AWS_SECRET_ACCESS_KEY,
        )

        mock_create_transfer_job = mock.MagicMock()
        mock_create_transfer_job.return_value = dict(name=TRANSFER_JOB_ID_0)
        mock_transfer_hook.return_value.create_transfer_job = mock_create_transfer_job

        operator = S3ToGCSOperator(
            task_id=TASK_ID,
            bucket=S3_BUCKET,
            prefix=S3_PREFIX,
            delimiter=S3_DELIMITER,
            gcp_conn_id=GCS_CONN_ID,
            dest_gcs=GCS_PATH_PREFIX,
            aws_conn_id=AWS_CONN_ID,
            replace=True,
            deferrable=True,
        )

        operator.hook = mock_s3_super_hook

        with pytest.raises(TaskDeferred) as exception_info:
            operator.execute(None)

        mock_s3_hook.assert_called_once_with(aws_conn_id=AWS_CONN_ID, verify=operator.verify)
        mock_s3_super_hook.list_keys.assert_called_once_with(
            bucket_name=S3_BUCKET, prefix=S3_PREFIX, delimiter=S3_DELIMITER, apply_wildcard=False
        )
        mock_create_transfer_job.assert_called_once()
        assert hasattr(exception_info.value, "trigger")
        trigger = exception_info.value.trigger
        assert trigger.project_id == PROJECT_ID
        assert trigger.job_names == [TRANSFER_JOB_ID_0]
        assert trigger.poll_interval == operator.poll_interval

        assert hasattr(exception_info.value, "method_name")
        assert exception_info.value.method_name == "execute_complete"

    @mock.patch("airflow.providers.google.cloud.transfers.s3_to_gcs.S3Hook")
    @mock.patch("airflow.providers.google.cloud.transfers.s3_to_gcs.GCSHook")
    def test_transfer_files_async(
        self,
        mock_s3_hook,
        mock_gcs_hook,
    ):
        mock_s3_hook.conn_config = mock.MagicMock(
            aws_access_key_id=AWS_ACCESS_KEY_ID,
            aws_secret_access_key=AWS_SECRET_ACCESS_KEY,
        )
        mock_gcs_hook.project_id = PROJECT_ID
        expected_job_names = [TRANSFER_JOB_ID_0]
        expected_method_name = "execute_complete"

        operator = S3ToGCSOperator(
            task_id=TASK_ID,
            bucket=S3_BUCKET,
            prefix=S3_PREFIX,
            delimiter=S3_DELIMITER,
            gcp_conn_id=GCS_CONN_ID,
            dest_gcs=GCS_PATH_PREFIX,
        )

        with mock.patch.object(operator, "submit_transfer_jobs") as mock_submit_transfer_jobs:
            mock_submit_transfer_jobs.return_value = expected_job_names
            with pytest.raises(TaskDeferred) as exception_info:
                operator.transfer_files_async(files=MOCK_FILES, gcs_hook=mock_gcs_hook, s3_hook=mock_s3_hook)

        mock_submit_transfer_jobs.assert_called_once_with(
            files=MOCK_FILES, gcs_hook=mock_gcs_hook, s3_hook=mock_s3_hook
        )

        assert hasattr(exception_info.value, "trigger")
        trigger = exception_info.value.trigger
        assert trigger.project_id == PROJECT_ID
        assert trigger.job_names == expected_job_names
        assert trigger.poll_interval == operator.poll_interval

        assert hasattr(exception_info.value, "method_name")
        assert exception_info.value.method_name == expected_method_name

    @pytest.mark.parametrize("invalid_poll_interval", [-5, 0])
    def test_init_error_polling_interval(self, invalid_poll_interval):
        operator = None
        expected_error_message = "Invalid value for poll_interval. Expected value greater than 0"
        with pytest.raises(ValueError, match=expected_error_message):
            operator = S3ToGCSOperator(
                task_id=TASK_ID,
                bucket=S3_BUCKET,
                prefix=S3_PREFIX,
                delimiter=S3_DELIMITER,
                gcp_conn_id=GCS_CONN_ID,
                dest_gcs=GCS_PATH_PREFIX,
                poll_interval=invalid_poll_interval,
            )
        assert operator is None

    def test_transfer_files_async_error_no_files(self):
        operator = S3ToGCSOperator(
            task_id=TASK_ID,
            bucket=S3_BUCKET,
            prefix=S3_PREFIX,
            delimiter=S3_DELIMITER,
            gcp_conn_id=GCS_CONN_ID,
            dest_gcs=GCS_PATH_PREFIX,
        )
        expected_error_message = "List of transferring files cannot be empty"
        with pytest.raises(ValueError, match=expected_error_message):
            operator.transfer_files_async(files=[], gcs_hook=mock.MagicMock(), s3_hook=mock.MagicMock())

    @pytest.mark.parametrize(
        ("file_names", "chunks", "expected_job_names"),
        [
            (MOCK_FILES, [MOCK_FILES], [TRANSFER_JOB_ID_0]),
            (
                [f"path/to/file{i}" for i in range(2000)],
                [
                    [f"path/to/file{i}" for i in range(1000)],
                    [f"path/to/file{i}" for i in range(1000, 2000)],
                ],
                TRANSFER_JOBS,
            ),
        ],
    )
    @mock.patch("airflow.providers.google.cloud.transfers.s3_to_gcs.S3Hook")
    @mock.patch("airflow.providers.google.cloud.transfers.s3_to_gcs.GCSHook")
    def test_submit_transfer_jobs(
        self,
        mock_s3_hook,
        mock_gcs_hook,
        file_names,
        chunks,
        expected_job_names,
    ):
        mock_s3_hook.conn_config = mock.MagicMock(
            aws_access_key_id=AWS_ACCESS_KEY_ID,
            aws_secret_access_key=AWS_SECRET_ACCESS_KEY,
        )
        mock_gcs_hook.project_id = PROJECT_ID
        operator = S3ToGCSOperator(
            task_id=TASK_ID,
            bucket=S3_BUCKET,
            prefix=S3_PREFIX,
            delimiter=S3_DELIMITER,
            gcp_conn_id=GCS_CONN_ID,
            dest_gcs=GCS_PATH_PREFIX,
        )

        now_time = utcnow()
        with time_machine.travel(now_time):
            with mock.patch.object(operator, "get_transfer_hook") as mock_get_transfer_hook:
                mock_create_transfer_job = mock.MagicMock(
                    side_effect=[dict(name=job_name) for job_name in expected_job_names]
                )
                mock_get_transfer_hook.return_value = mock.MagicMock(
                    create_transfer_job=mock_create_transfer_job
                )
                job_names = operator.submit_transfer_jobs(
                    files=file_names,
                    gcs_hook=mock_gcs_hook,
                    s3_hook=mock_s3_hook,
                )

        mock_get_transfer_hook.assert_called_once()
        mock_create_transfer_job.assert_called()
        assert job_names == expected_job_names

    @mock.patch(
        "airflow.providers.google.cloud.transfers.s3_to_gcs.S3ToGCSOperator.log", new_callable=PropertyMock
    )
    def test_execute_complete_success(self, mock_log):
        expected_event_message = "Event message (success)"
        event = {
            "status": "success",
            "message": expected_event_message,
        }
        operator = S3ToGCSOperator(task_id=TASK_ID, bucket=S3_BUCKET)
        operator.execute_complete(context=mock.MagicMock(), event=event)

        mock_log.return_value.info.assert_called_once_with(
            "%s completed with response %s ", TASK_ID, event["message"]
        )

    @mock.patch(
        "airflow.providers.google.cloud.transfers.s3_to_gcs.S3ToGCSOperator.log", new_callable=PropertyMock
    )
    def test_execute_complete_error(self, mock_log):
        expected_event_message = "Event error message"
        event = {
            "status": "error",
            "message": expected_event_message,
        }
        operator = S3ToGCSOperator(task_id=TASK_ID, bucket=S3_BUCKET)
        with pytest.raises(AirflowException, match=expected_event_message):
            operator.execute_complete(context=mock.MagicMock(), event=event)

        mock_log.return_value.info.assert_not_called()

    @pytest.mark.db_test
    def test_get_transfer_hook(self):
        operator = S3ToGCSOperator(
            task_id=TASK_ID,
            bucket=S3_BUCKET,
            prefix=S3_PREFIX,
            delimiter=S3_DELIMITER,
            gcp_conn_id=GCS_CONN_ID,
            dest_gcs=GCS_PATH_PREFIX,
            google_impersonation_chain=IMPERSONATION_CHAIN,
        )
        transfer_hook = operator.get_transfer_hook()

        assert isinstance(transfer_hook, CloudDataTransferServiceHook)
        assert transfer_hook.gcp_conn_id == GCS_CONN_ID
        assert transfer_hook.impersonation_chain == IMPERSONATION_CHAIN



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import json
import os
from unittest import mock

from airflow.models import Connection

GCP_PROJECT_ID_HOOK_UNIT_TEST = "example-project"


def mock_base_gcp_hook_default_project_id(
    self,
    gcp_conn_id="google_cloud_default",
    impersonation_chain=None,
):
    self.standard_extras_list = {"project": GCP_PROJECT_ID_HOOK_UNIT_TEST}
    self._conn = gcp_conn_id
    self.impersonation_chain = impersonation_chain
    self._client = None
    self._conn = None
    self._cached_credentials = None
    self._cached_project_id = None


def mock_base_gcp_hook_no_default_project_id(
    self,
    gcp_conn_id="google_cloud_default",
    impersonation_chain=None,
):
    self.standard_extras_list = {}
    self._conn = gcp_conn_id
    self.impersonation_chain = impersonation_chain
    self._client = None
    self._conn = None
    self._cached_credentials = None
    self._cached_project_id = None


if os.environ.get("_AIRFLOW_SKIP_DB_TESTS") == "true":
    GCP_CONNECTION_WITH_PROJECT_ID = None
    GCP_CONNECTION_WITHOUT_PROJECT_ID = None

else:
    GCP_CONNECTION_WITH_PROJECT_ID = Connection(extra=json.dumps({"project": GCP_PROJECT_ID_HOOK_UNIT_TEST}))
    GCP_CONNECTION_WITHOUT_PROJECT_ID = Connection(extra=json.dumps({}))


def get_open_mock():
    mck = mock.mock_open()
    open_module = "builtins"
    return mck, open_module



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""
Example Airflow DAG that translates text in Google Cloud Translate using V3 API version
service in the Google Cloud.
"""

from __future__ import annotations

import os
from datetime import datetime

from airflow.models.dag import DAG
from airflow.providers.google.cloud.operators.gcs import GCSCreateBucketOperator, GCSDeleteBucketOperator
from airflow.providers.google.cloud.operators.translate import (
    TranslateCreateDatasetOperator,
    TranslateCreateModelOperator,
    TranslateDeleteDatasetOperator,
    TranslateDeleteModelOperator,
    TranslateImportDataOperator,
    TranslateModelsListOperator,
    TranslateTextOperator,
)
from airflow.providers.google.cloud.transfers.gcs_to_gcs import GCSToGCSOperator

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

DAG_ID = "gcp_translate_automl_native_model"
PROJECT_ID = os.environ.get("SYSTEM_TESTS_GCP_PROJECT", "default")
ENV_ID = os.environ.get("SYSTEM_TESTS_ENV_ID", "default")
REGION = "us-central1"
RESOURCE_DATA_BUCKET = "airflow-system-tests-resources"
DATA_SAMPLE_GCS_BUCKET_NAME = f"bucket_{DAG_ID}_{ENV_ID}".replace("_", "-")
DATA_FILE_NAME = "import_en-es.tsv"
RESOURCE_PATH = f"V3_translate/create_ds/import_data/{DATA_FILE_NAME}"
COPY_DATA_PATH = f"gs://{RESOURCE_DATA_BUCKET}/V3_translate/create_ds/import_data/{DATA_FILE_NAME}"
DST_PATH = f"translate/import/{DATA_FILE_NAME}"
DATASET_DATA_PATH = f"gs://{DATA_SAMPLE_GCS_BUCKET_NAME}/{DST_PATH}"
DATASET = {
    "display_name": f"ds_native_{DAG_ID}_{ENV_ID}",
    "source_language_code": "en",
    "target_language_code": "es",
}


with DAG(
    DAG_ID,
    schedule="@once",  # Override to match your needs
    start_date=datetime(2025, 1, 1),
    catchup=False,
    tags=[
        "example",
        "translate_native_model",
    ],
) as dag:
    create_bucket = GCSCreateBucketOperator(
        task_id="create_bucket",
        bucket_name=DATA_SAMPLE_GCS_BUCKET_NAME,
        storage_class="REGIONAL",
        location=REGION,
    )
    copy_dataset_source_tsv = GCSToGCSOperator(
        task_id="copy_dataset_file",
        source_bucket=RESOURCE_DATA_BUCKET,
        source_object=RESOURCE_PATH,
        destination_bucket=DATA_SAMPLE_GCS_BUCKET_NAME,
        destination_object=DST_PATH,
    )

    create_dataset_op = TranslateCreateDatasetOperator(
        task_id="translate_v3_ds_create",
        dataset=DATASET,
        project_id=PROJECT_ID,
        location=REGION,
    )

    import_ds_data_op = TranslateImportDataOperator(
        task_id="translate_v3_ds_import_data",
        dataset_id=create_dataset_op.output["dataset_id"],
        input_config={
            "input_files": [{"usage": "UNASSIGNED", "gcs_source": {"input_uri": DATASET_DATA_PATH}}]
        },
        project_id=PROJECT_ID,
        location=REGION,
    )

    # [START howto_operator_translate_automl_create_model]
    create_model = TranslateCreateModelOperator(
        task_id="translate_v3_model_create",
        display_name=f"native_model_{ENV_ID}"[:32].replace("-", "_"),
        dataset_id=create_dataset_op.output["dataset_id"],
        project_id=PROJECT_ID,
        location=REGION,
    )
    # [END howto_operator_translate_automl_create_model]

    # [START howto_operator_translate_automl_list_models]
    list_models = TranslateModelsListOperator(
        task_id="translate_v3_list_models",
        project_id=PROJECT_ID,
        location=REGION,
    )
    # [END howto_operator_translate_automl_list_models]

    model_id = create_model.output["model_id"]

    translate_text_with_model = TranslateTextOperator(
        task_id="translate_v3_op",
        contents=["Hello!", "Can I have a cup of coffee, please?"],
        # AutoML model format
        model=f"projects/{PROJECT_ID}/locations/{REGION}/models/{model_id}",
        source_language_code="en",
        target_language_code="es",
        location=REGION,
    )

    # [START howto_operator_translate_automl_delete_model]
    delete_model = TranslateDeleteModelOperator(
        task_id="translate_v3_automl_delete_model",
        model_id=model_id,
        project_id=PROJECT_ID,
        location=REGION,
    )
    # [END howto_operator_translate_automl_delete_model]

    delete_ds_op = TranslateDeleteDatasetOperator(
        task_id="translate_v3_ds_delete",
        dataset_id=create_dataset_op.output["dataset_id"],
        project_id=PROJECT_ID,
        location=REGION,
    )
    # [END howto_operator_translate_automl_delete_dataset]

    delete_bucket = GCSDeleteBucketOperator(
        task_id="delete_bucket",
        bucket_name=DATA_SAMPLE_GCS_BUCKET_NAME,
        trigger_rule=TriggerRule.ALL_DONE,
    )

    (
        # TEST SETUP
        [create_bucket >> copy_dataset_source_tsv]
        >> create_dataset_op
        >> import_ds_data_op
        # TEST BODY
        >> create_model
        >> list_models
        >> translate_text_with_model
        >> delete_model
        # TEST TEARDOWN
        >> delete_ds_op
        >> delete_bucket
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()


from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module contains Google Vertex AI Generative AI operators."""

from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING, Any, Literal

from google.api_core import exceptions

from airflow.exceptions import AirflowException, AirflowProviderDeprecationWarning
from airflow.providers.google.cloud.hooks.vertex_ai.generative_model import (
    ExperimentRunHook,
    GenerativeModelHook,
)
from airflow.providers.google.cloud.operators.cloud_base import GoogleCloudBaseOperator
from airflow.providers.google.common.deprecated import deprecated

if TYPE_CHECKING:
    from airflow.providers.common.compat.sdk import Context


@deprecated(
    planned_removal_date="January 3, 2026",
    use_instead="airflow.providers.google.cloud.operators.gen_ai.generative_model.GenAIGenerateEmbeddingsOperator",
    category=AirflowProviderDeprecationWarning,
)
class TextEmbeddingModelGetEmbeddingsOperator(GoogleCloudBaseOperator):
    """
    Uses the Vertex AI Embeddings API to generate embeddings based on prompt.

    :param project_id: Required. The ID of the Google Cloud project that the
        service belongs to (templated).
    :param location: Required. The ID of the Google Cloud location that the
        service belongs to (templated).
    :param prompt: Required. Inputs or queries that a user or a program gives
        to the Vertex AI Generative Model API, in order to elicit a specific response (templated).
    :param pretrained_model: Required. Model, optimized for performing text embeddings.
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields = ("location", "project_id", "impersonation_chain", "prompt", "pretrained_model")

    def __init__(
        self,
        *,
        project_id: str,
        location: str,
        prompt: str,
        pretrained_model: str,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.project_id = project_id
        self.location = location
        self.prompt = prompt
        self.pretrained_model = pretrained_model
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

    def execute(self, context: Context):
        self.hook = GenerativeModelHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )

        self.log.info("Generating text embeddings")
        response = self.hook.text_embedding_model_get_embeddings(
            project_id=self.project_id,
            location=self.location,
            prompt=self.prompt,
            pretrained_model=self.pretrained_model,
        )

        self.log.info("Model response: %s", response)
        context["ti"].xcom_push(key="model_response", value=response)

        return response


@deprecated(
    planned_removal_date="January 3, 2026",
    use_instead="airflow.providers.google.cloud.operators.gen_ai.generative_model.GenAIGenerateContentOperator",
    category=AirflowProviderDeprecationWarning,
)
class GenerativeModelGenerateContentOperator(GoogleCloudBaseOperator):
    """
    Use the Vertex AI Gemini Pro foundation model to generate content.

    :param project_id: Required. The ID of the Google Cloud project that the
        service belongs to (templated).
    :param location: Required. The ID of the Google Cloud location that the
        service belongs to (templated).
    :param contents: Required. The multi-part content of a message that a user or a program
        gives to the generative model, in order to elicit a specific response.
    :param generation_config: Optional. Generation configuration settings.
    :param safety_settings: Optional. Per request settings for blocking unsafe content.
    :param tools: Optional. A list of tools available to the model during evaluation, such as a data store.
    :param system_instruction: Optional. An instruction given to the model to guide its behavior.
    :param pretrained_model: Required. The name of the model to use for content generation,
        which can be a text-only or multimodal model. For example, `gemini-pro` or
        `gemini-pro-vision`.
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields = ("location", "project_id", "impersonation_chain", "contents", "pretrained_model")

    def __init__(
        self,
        *,
        project_id: str,
        location: str,
        contents: list,
        tools: list | None = None,
        generation_config: dict | None = None,
        safety_settings: dict | None = None,
        system_instruction: str | None = None,
        pretrained_model: str,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.project_id = project_id
        self.location = location
        self.contents = contents
        self.tools = tools
        self.generation_config = generation_config
        self.safety_settings = safety_settings
        self.system_instruction = system_instruction
        self.pretrained_model = pretrained_model
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

    def execute(self, context: Context):
        self.hook = GenerativeModelHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )
        response = self.hook.generative_model_generate_content(
            project_id=self.project_id,
            location=self.location,
            contents=self.contents,
            tools=self.tools,
            generation_config=self.generation_config,
            safety_settings=self.safety_settings,
            system_instruction=self.system_instruction,
            pretrained_model=self.pretrained_model,
        )

        self.log.info("Model response: %s", response)
        context["ti"].xcom_push(key="model_response", value=response)

        return response


@deprecated(
    planned_removal_date="January 3, 2026",
    use_instead="airflow.providers.google.cloud.operators.gen_ai.generative_model.GenAISupervisedFineTuningTrainOperator",
    category=AirflowProviderDeprecationWarning,
)
class SupervisedFineTuningTrainOperator(GoogleCloudBaseOperator):
    """
    Use the Supervised Fine Tuning API to create a tuning job.

    :param project_id: Required. The ID of the Google Cloud project that the
        service belongs to.
    :param location: Required. The ID of the Google Cloud location that the service belongs to.
    :param source_model: Required. A pre-trained model optimized for performing natural
        language tasks such as classification, summarization, extraction, content
        creation, and ideation.
    :param train_dataset: Required. Cloud Storage URI of your training dataset. The dataset
        must be formatted as a JSONL file. For best results, provide at least 100 to 500 examples.
    :param tuned_model_display_name: Optional. Display name of the TunedModel. The name can be up
        to 128 characters long and can consist of any UTF-8 characters.
    :param validation_dataset: Optional. Cloud Storage URI of your training dataset. The dataset must be
        formatted as a JSONL file. For best results, provide at least 100 to 500 examples.
    :param epochs: Optional. To optimize performance on a specific dataset, try using a higher
        epoch value. Increasing the number of epochs might improve results. However, be cautious
        about over-fitting, especially when dealing with small datasets. If over-fitting occurs,
        consider lowering the epoch number.
    :param adapter_size: Optional. Adapter size for tuning.
    :param learning_multiplier_rate: Optional. Multiplier for adjusting the default learning rate.
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields = (
        "location",
        "project_id",
        "impersonation_chain",
        "train_dataset",
        "validation_dataset",
        "source_model",
    )

    def __init__(
        self,
        *,
        project_id: str,
        location: str,
        source_model: str,
        train_dataset: str,
        tuned_model_display_name: str | None = None,
        validation_dataset: str | None = None,
        epochs: int | None = None,
        adapter_size: Literal[1, 4, 8, 16] | None = None,
        learning_rate_multiplier: float | None = None,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.project_id = project_id
        self.location = location
        self.source_model = source_model
        self.train_dataset = train_dataset
        self.tuned_model_display_name = tuned_model_display_name
        self.validation_dataset = validation_dataset
        self.epochs = epochs
        self.adapter_size = adapter_size
        self.learning_rate_multiplier = learning_rate_multiplier
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

    def execute(self, context: Context):
        self.hook = GenerativeModelHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )
        response = self.hook.supervised_fine_tuning_train(
            project_id=self.project_id,
            location=self.location,
            source_model=self.source_model,
            train_dataset=self.train_dataset,
            validation_dataset=self.validation_dataset,
            epochs=self.epochs,
            adapter_size=self.adapter_size,
            learning_rate_multiplier=self.learning_rate_multiplier,
            tuned_model_display_name=self.tuned_model_display_name,
        )

        self.log.info("Tuned Model Name: %s", response.tuned_model_name)
        self.log.info("Tuned Model Endpoint Name: %s", response.tuned_model_endpoint_name)

        context["ti"].xcom_push(key="tuned_model_name", value=response.tuned_model_name)
        context["ti"].xcom_push(key="tuned_model_endpoint_name", value=response.tuned_model_endpoint_name)

        result = {
            "tuned_model_name": response.tuned_model_name,
            "tuned_model_endpoint_name": response.tuned_model_endpoint_name,
        }

        return result


@deprecated(
    planned_removal_date="January 3, 2026",
    use_instead="airflow.providers.google.cloud.operators.gen_ai.generative_model.GenAICountTokensOperator",
    category=AirflowProviderDeprecationWarning,
)
class CountTokensOperator(GoogleCloudBaseOperator):
    """
    Use the Vertex AI Count Tokens API to calculate the number of input tokens before sending a request to the Gemini API.

    :param project_id: Required. The ID of the Google Cloud project that the
        service belongs to (templated).
    :param location: Required. The ID of the Google Cloud location that the
        service belongs to (templated).
    :param contents: Required. The multi-part content of a message that a user or a program
        gives to the generative model, in order to elicit a specific response.
    :param pretrained_model: Required. Model, supporting prompts with text-only input,
        including natural language tasks, multi-turn text and code chat,
        and code generation. It can output text and code.
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields = ("location", "project_id", "impersonation_chain", "contents", "pretrained_model")

    def __init__(
        self,
        *,
        project_id: str,
        location: str,
        contents: list,
        pretrained_model: str,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.project_id = project_id
        self.location = location
        self.contents = contents
        self.pretrained_model = pretrained_model
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

    def execute(self, context: Context):
        self.hook = GenerativeModelHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )
        response = self.hook.count_tokens(
            project_id=self.project_id,
            location=self.location,
            contents=self.contents,
            pretrained_model=self.pretrained_model,
        )

        self.log.info("Total tokens: %s", response.total_tokens)
        self.log.info("Total billable characters: %s", response.total_billable_characters)

        context["ti"].xcom_push(key="total_tokens", value=response.total_tokens)
        context["ti"].xcom_push(key="total_billable_characters", value=response.total_billable_characters)


class RunEvaluationOperator(GoogleCloudBaseOperator):
    """
    Use the Rapid Evaluation API to evaluate a model.

    :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
    :param location: Required. The ID of the Google Cloud location that the service belongs to.
    :param pretrained_model: Required. A pre-trained model optimized for performing natural
        language tasks such as classification, summarization, extraction, content
        creation, and ideation.
    :param eval_dataset: Required. A fixed dataset for evaluating a model against. Adheres to Rapid Evaluation API.
    :param metrics: Required. A list of evaluation metrics to be used in the experiment. Adheres to Rapid Evaluation API.
    :param experiment_name: Required. The name of the evaluation experiment.
    :param experiment_run_name: Required. The specific run name or ID for this experiment.
    :param prompt_template: Required. The template used to format the model's prompts during evaluation. Adheres to Rapid Evaluation API.
    :param generation_config: Optional. A dictionary containing generation parameters for the model.
    :param safety_settings: Optional. A dictionary specifying harm category thresholds for blocking model outputs.
    :param system_instruction: Optional. An instruction given to the model to guide its behavior.
    :param tools: Optional. A list of tools available to the model during evaluation, such as a data store.
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields = (
        "location",
        "project_id",
        "impersonation_chain",
        "pretrained_model",
        "eval_dataset",
        "prompt_template",
        "experiment_name",
        "experiment_run_name",
    )

    def __init__(
        self,
        *,
        project_id: str,
        location: str,
        pretrained_model: str,
        eval_dataset: dict,
        metrics: list,
        experiment_name: str,
        experiment_run_name: str,
        prompt_template: str,
        generation_config: dict | None = None,
        safety_settings: dict | None = None,
        system_instruction: str | None = None,
        tools: list | None = None,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        self.project_id = project_id
        self.location = location
        self.pretrained_model = pretrained_model
        self.eval_dataset = eval_dataset
        self.metrics = metrics
        self.experiment_name = experiment_name
        self.experiment_run_name = experiment_run_name
        self.prompt_template = prompt_template
        self.generation_config = generation_config
        self.safety_settings = safety_settings
        self.system_instruction = system_instruction
        self.tools = tools
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

    def execute(self, context: Context):
        self.hook = GenerativeModelHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )
        response = self.hook.run_evaluation(
            project_id=self.project_id,
            location=self.location,
            pretrained_model=self.pretrained_model,
            eval_dataset=self.eval_dataset,
            metrics=self.metrics,
            experiment_name=self.experiment_name,
            experiment_run_name=self.experiment_run_name,
            prompt_template=self.prompt_template,
            generation_config=self.generation_config,
            safety_settings=self.safety_settings,
            system_instruction=self.system_instruction,
            tools=self.tools,
        )

        return response.summary_metrics


@deprecated(
    planned_removal_date="January 3, 2026",
    use_instead="airflow.providers.google.cloud.operators.gen_ai.generative_model.GenAICreateCachedContentOperator",
    category=AirflowProviderDeprecationWarning,
)
class CreateCachedContentOperator(GoogleCloudBaseOperator):
    """
    Create CachedContent to reduce the cost of requests that contain repeat content with high input token counts.

    :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
    :param location: Required. The ID of the Google Cloud location that the service belongs to.
    :param model_name: Required. The name of the publisher model to use for cached content.
    :param system_instruction: Developer set system instruction.
    :param contents: The content to cache.
    :param ttl_hours: The TTL for this resource in hours. The expiration time is computed: now + TTL.
        Defaults to one hour.
    :param display_name: The user-generated meaningful display name of the cached content
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields = (
        "location",
        "project_id",
        "impersonation_chain",
        "model_name",
        "contents",
        "system_instruction",
    )

    def __init__(
        self,
        *,
        project_id: str,
        location: str,
        model_name: str,
        system_instruction: Any | None = None,
        contents: list[Any] | None = None,
        ttl_hours: float = 1,
        display_name: str | None = None,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        self.project_id = project_id
        self.location = location
        self.model_name = model_name
        self.system_instruction = system_instruction
        self.contents = contents
        self.ttl_hours = ttl_hours
        self.display_name = display_name
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

    def execute(self, context: Context):
        self.hook = GenerativeModelHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )

        cached_content_name = self.hook.create_cached_content(
            project_id=self.project_id,
            location=self.location,
            model_name=self.model_name,
            system_instruction=self.system_instruction,
            contents=self.contents,
            ttl_hours=self.ttl_hours,
            display_name=self.display_name,
        )

        self.log.info("Cached Content Name: %s", cached_content_name)

        return cached_content_name


@deprecated(
    planned_removal_date="January 3, 2026",
    use_instead="airflow.providers.google.cloud.operators.gen_ai.generative_model.GenAIGenerateContentOperator",
    category=AirflowProviderDeprecationWarning,
)
class GenerateFromCachedContentOperator(GoogleCloudBaseOperator):
    """
    Generate a response from CachedContent.

    :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
    :param location: Required. The ID of the Google Cloud location that the service belongs to.
    :param cached_content_name: Required. The name of the cached content resource.
    :param contents: Required. The multi-part content of a message that a user or a program
        gives to the generative model, in order to elicit a specific response.
    :param generation_config: Optional. Generation configuration settings.
    :param safety_settings: Optional. Per request settings for blocking unsafe content.
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields = (
        "location",
        "project_id",
        "impersonation_chain",
        "cached_content_name",
        "contents",
    )

    def __init__(
        self,
        *,
        project_id: str,
        location: str,
        cached_content_name: str,
        contents: list,
        generation_config: dict | None = None,
        safety_settings: dict | None = None,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        self.project_id = project_id
        self.location = location
        self.cached_content_name = cached_content_name
        self.contents = contents
        self.generation_config = generation_config
        self.safety_settings = safety_settings
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

    def execute(self, context: Context):
        self.hook = GenerativeModelHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )
        cached_content_text = self.hook.generate_from_cached_content(
            project_id=self.project_id,
            location=self.location,
            cached_content_name=self.cached_content_name,
            contents=self.contents,
            generation_config=self.generation_config,
            safety_settings=self.safety_settings,
        )

        self.log.info("Cached Content Response: %s", cached_content_text)

        return cached_content_text


@deprecated(
    planned_removal_date="January 3, 2026",
    use_instead="airflow.providers.google.cloud.operators.vertex_ai.experiment_service.DeleteExperimentRunOperator",
    category=AirflowProviderDeprecationWarning,
)
class DeleteExperimentRunOperator(GoogleCloudBaseOperator):
    """
    Use the Rapid Evaluation API to evaluate a model.

    :param project_id: Required. The ID of the Google Cloud project that the service belongs to.
    :param location: Required. The ID of the Google Cloud location that the service belongs to.
    :param experiment_name: Required. The name of the evaluation experiment.
    :param experiment_run_name: Required. The specific run name or ID for this experiment.
    :param gcp_conn_id: The connection ID to use connecting to Google Cloud.
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields = (
        "location",
        "project_id",
        "impersonation_chain",
        "experiment_name",
        "experiment_run_name",
    )

    def __init__(
        self,
        *,
        project_id: str,
        location: str,
        experiment_name: str,
        experiment_run_name: str,
        gcp_conn_id: str = "google_cloud_default",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.project_id = project_id
        self.location = location
        self.experiment_name = experiment_name
        self.experiment_run_name = experiment_run_name
        self.gcp_conn_id = gcp_conn_id
        self.impersonation_chain = impersonation_chain

    def execute(self, context: Context) -> None:
        self.hook = ExperimentRunHook(
            gcp_conn_id=self.gcp_conn_id,
            impersonation_chain=self.impersonation_chain,
        )

        try:
            self.hook.delete_experiment_run(
                project_id=self.project_id,
                location=self.location,
                experiment_name=self.experiment_name,
                experiment_run_name=self.experiment_run_name,
            )
        except exceptions.NotFound:
            raise AirflowException(f"Experiment Run with name {self.experiment_run_name} not found")

        self.log.info("Deleted experiment run: %s", self.experiment_run_name)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING

from airflow.exceptions import AirflowException
from airflow.providers.google.common.hooks.base_google import PROVIDE_PROJECT_ID
from airflow.providers.google.firebase.hooks.firestore import CloudFirestoreHook
from airflow.providers.google.version_compat import BaseOperator

if TYPE_CHECKING:
    from airflow.providers.common.compat.sdk import Context


class CloudFirestoreExportDatabaseOperator(BaseOperator):
    """
    Export documents from Google Cloud Firestore to another storage system, such as Google Cloud Storage.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:CloudFirestoreExportDatabaseOperator`

    :param database_id: The Database ID.
    :param body: The request body.
        See:
        https://firebase.google.com/docs/firestore/reference/rest/v1beta1/projects.databases/exportDocuments
    :param project_id: ID of the Google Cloud project if None then
        default project_id is used.
    :param gcp_conn_id: The connection ID to use to connect to Google Cloud.
    :param api_version: API version used (for example v1 or v1beta1).
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    """

    template_fields: Sequence[str] = (
        "body",
        "gcp_conn_id",
        "api_version",
        "impersonation_chain",
    )

    def __init__(
        self,
        *,
        body: dict,
        database_id: str = "(default)",
        project_id: str = PROVIDE_PROJECT_ID,
        gcp_conn_id: str = "google_cloud_default",
        api_version: str = "v1",
        impersonation_chain: str | Sequence[str] | None = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.database_id = database_id
        self.body = body
        self.project_id = project_id
        self.gcp_conn_id = gcp_conn_id
        self.api_version = api_version
        self._validate_inputs()
        self.impersonation_chain = impersonation_chain

    def _validate_inputs(self) -> None:
        if not self.body:
            raise AirflowException("The required parameter 'body' is missing")

    def execute(self, context: Context):
        hook = CloudFirestoreHook(
            gcp_conn_id=self.gcp_conn_id,
            api_version=self.api_version,
            impersonation_chain=self.impersonation_chain,
        )
        return hook.export_documents(database_id=self.database_id, body=self.body, project_id=self.project_id)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""
Example Airflow DAG that creates and deletes Queues and creates, gets, lists,
runs and deletes Tasks in the Google Cloud Tasks service in the Google Cloud.
"""

from __future__ import annotations

import os
from datetime import datetime, timedelta

from google.api_core.retry import Retry
from google.cloud.tasks_v2.types import Queue
from google.protobuf import timestamp_pb2

from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS

if AIRFLOW_V_3_0_PLUS:
    from airflow.sdk import task
else:
    # Airflow 2 path
    from airflow.decorators import task  # type: ignore[attr-defined,no-redef]
from airflow.models.baseoperator import chain
from airflow.models.dag import DAG
from airflow.providers.google.cloud.operators.tasks import (
    CloudTasksQueueCreateOperator,
    CloudTasksQueueDeleteOperator,
    CloudTasksTaskCreateOperator,
    CloudTasksTaskDeleteOperator,
    CloudTasksTaskGetOperator,
    CloudTasksTaskRunOperator,
    CloudTasksTasksListOperator,
)

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

ENV_ID = os.environ.get("SYSTEM_TESTS_ENV_ID", "default")
DAG_ID = "cloud_tasks_tasks"

timestamp = timestamp_pb2.Timestamp()
timestamp.FromDatetime(datetime.now() + timedelta(hours=12))

LOCATION = "us-central1"
# queue cannot use recent names even if queue was removed
QUEUE_ID = f"queue-{ENV_ID}-{DAG_ID.replace('_', '-')}"
TASK_NAME = "task-to-run"
TASK = {
    "http_request": {
        "http_method": "POST",
        "url": "http://www.example.com/example",
        "body": b"",
    },
    "schedule_time": timestamp,
}

with DAG(
    dag_id=DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    catchup=False,
    tags=["example", "tasks"],
) as dag:

    @task(task_id="random_string")
    def generate_random_string():
        """
        Generate random string for queue and task names.
        Queue name cannot be repeated in preceding 7 days and
        task name in the last 1 hour.
        """
        import random
        import string

        return "".join(random.choices(string.ascii_uppercase + string.digits, k=8))

    random_string = generate_random_string()

    create_queue = CloudTasksQueueCreateOperator(
        location=LOCATION,
        task_queue=Queue(stackdriver_logging_config=dict(sampling_ratio=0.5)),
        queue_name=QUEUE_ID + "{{ task_instance.xcom_pull(task_ids='random_string') }}",
        retry=Retry(maximum=10.0),
        timeout=5,
        task_id="create_queue",
    )

    delete_queue = CloudTasksQueueDeleteOperator(
        location=LOCATION,
        queue_name=QUEUE_ID + "{{ task_instance.xcom_pull(task_ids='random_string') }}",
        task_id="delete_queue",
    )
    delete_queue.trigger_rule = TriggerRule.ALL_DONE

    # [START create_task]
    create_task = CloudTasksTaskCreateOperator(
        location=LOCATION,
        queue_name=QUEUE_ID + "{{ task_instance.xcom_pull(task_ids='random_string') }}",
        task=TASK,
        task_name=TASK_NAME + "{{ task_instance.xcom_pull(task_ids='random_string') }}",
        retry=Retry(maximum=10.0),
        timeout=5,
        task_id="create_task_to_run",
    )
    # [END create_task]

    # [START tasks_get]
    tasks_get = CloudTasksTaskGetOperator(
        location=LOCATION,
        queue_name=QUEUE_ID + "{{ task_instance.xcom_pull(task_ids='random_string') }}",
        task_name=TASK_NAME + "{{ task_instance.xcom_pull(task_ids='random_string') }}",
        task_id="tasks_get",
    )
    # [END tasks_get]

    # [START run_task]
    run_task = CloudTasksTaskRunOperator(
        location=LOCATION,
        queue_name=QUEUE_ID + "{{ task_instance.xcom_pull(task_ids='random_string') }}",
        task_name=TASK_NAME + "{{ task_instance.xcom_pull(task_ids='random_string') }}",
        retry=Retry(maximum=10.0),
        task_id="run_task",
    )
    # [END run_task]

    # [START list_tasks]
    list_tasks = CloudTasksTasksListOperator(
        location=LOCATION,
        queue_name=QUEUE_ID + "{{ task_instance.xcom_pull(task_ids='random_string') }}",
        task_id="list_tasks",
    )
    # [END list_tasks]

    # [START delete_task]
    delete_task = CloudTasksTaskDeleteOperator(
        location=LOCATION,
        queue_name=QUEUE_ID + "{{ task_instance.xcom_pull(task_ids='random_string') }}",
        task_name=TASK_NAME + "{{ task_instance.xcom_pull(task_ids='random_string') }}",
        task_id="delete_task",
    )
    # [END delete_task]

    chain(
        random_string, create_queue, create_task, tasks_get, list_tasks, run_task, delete_task, delete_queue
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

import pytest
from google.api_core.gapic_v1.method import DEFAULT
from google.cloud.bigquery_datatransfer_v1 import StartManualTransferRunsResponse, TransferConfig, TransferRun

from airflow.providers.google.cloud.operators.bigquery_dts import (
    BigQueryCreateDataTransferOperator,
    BigQueryDataTransferServiceStartTransferRunsOperator,
    BigQueryDeleteDataTransferConfigOperator,
)

PROJECT_ID = "id"

TRANSFER_CONFIG = {
    "data_source_id": "google_cloud_storage",
    "destination_dataset_id": "example_dataset",
    "params": {},
    "display_name": "example-transfer",
    "disabled": False,
    "data_refresh_window_days": 0,
    "schedule": "first sunday of quarter 00:00",
}

TRANSFER_CONFIG_ID = "id1234"

TRANSFER_CONFIG_NAME = "projects/123abc/locations/321cba/transferConfig/1a2b3c"
RUN_NAME = "projects/123abc/locations/321cba/transferConfig/1a2b3c/runs/123"
transfer_config = TransferConfig(
    name=TRANSFER_CONFIG_NAME, params={"secret_access_key": "AIRFLOW_KEY", "access_key_id": "AIRFLOW_KEY_ID"}
)


class TestBigQueryCreateDataTransferOperator:
    @mock.patch(
        "airflow.providers.google.cloud.operators.bigquery_dts.BiqQueryDataTransferServiceHook",
        **{"return_value.create_transfer_config.return_value": transfer_config},
    )
    def test_execute(self, mock_hook):
        op = BigQueryCreateDataTransferOperator(
            transfer_config=TRANSFER_CONFIG, project_id=PROJECT_ID, task_id="id"
        )
        ti = mock.MagicMock()

        return_value = op.execute({"ti": ti, "task": ti.task})

        mock_hook.return_value.create_transfer_config.assert_called_once_with(
            authorization_code=None,
            metadata=(),
            transfer_config=TRANSFER_CONFIG,
            project_id=PROJECT_ID,
            retry=DEFAULT,
            timeout=None,
        )
        ti.xcom_push.assert_called_with(key="transfer_config_id", value="1a2b3c")

        assert "secret_access_key" not in return_value.get("params", {})
        assert "access_key_id" not in return_value.get("params", {})


class TestBigQueryDeleteDataTransferConfigOperator:
    @mock.patch("airflow.providers.google.cloud.operators.bigquery_dts.BiqQueryDataTransferServiceHook")
    def test_execute(self, mock_hook):
        op = BigQueryDeleteDataTransferConfigOperator(
            transfer_config_id=TRANSFER_CONFIG_ID, task_id="id", project_id=PROJECT_ID
        )
        op.execute(None)
        mock_hook.return_value.delete_transfer_config.assert_called_once_with(
            metadata=(),
            transfer_config_id=TRANSFER_CONFIG_ID,
            project_id=PROJECT_ID,
            retry=DEFAULT,
            timeout=None,
        )


class TestBigQueryDataTransferServiceStartTransferRunsOperator:
    OPERATOR_MODULE_PATH = "airflow.providers.google.cloud.operators.bigquery_dts"

    @mock.patch(
        f"{OPERATOR_MODULE_PATH}.BigQueryDataTransferServiceStartTransferRunsOperator"
        f"._wait_for_transfer_to_be_done",
        mock.MagicMock(),
    )
    @mock.patch(
        f"{OPERATOR_MODULE_PATH}.BiqQueryDataTransferServiceHook",
        **{
            "return_value.start_manual_transfer_runs.return_value": StartManualTransferRunsResponse(
                runs=[TransferRun(name=RUN_NAME)]
            ),
        },
    )
    def test_execute(self, mock_hook):
        op = BigQueryDataTransferServiceStartTransferRunsOperator(
            transfer_config_id=TRANSFER_CONFIG_ID, task_id="id", project_id=PROJECT_ID
        )
        ti = mock.MagicMock()

        op.execute({"ti": ti, "task": ti.task})

        mock_hook.return_value.start_manual_transfer_runs.assert_called_once_with(
            transfer_config_id=TRANSFER_CONFIG_ID,
            project_id=PROJECT_ID,
            requested_time_range=None,
            requested_run_time=None,
            metadata=(),
            retry=DEFAULT,
            timeout=None,
        )
        ti.xcom_push.assert_called_with(key="run_id", value="123")

    @mock.patch(
        f"{OPERATOR_MODULE_PATH}.BiqQueryDataTransferServiceHook",
        **{
            "return_value.start_manual_transfer_runs.return_value": StartManualTransferRunsResponse(
                runs=[TransferRun(name=RUN_NAME)]
            ),
        },
    )
    @mock.patch(f"{OPERATOR_MODULE_PATH}.BigQueryDataTransferServiceStartTransferRunsOperator.defer")
    def test_defer_mode(self, _, defer_method):
        op = BigQueryDataTransferServiceStartTransferRunsOperator(
            transfer_config_id=TRANSFER_CONFIG_ID,
            task_id="id",
            project_id=PROJECT_ID,
            deferrable=True,
        )
        ti = mock.MagicMock()

        op.execute({"ti": ti, "task": ti.task})

        defer_method.assert_called_once()

    @pytest.mark.parametrize(
        ("data_source_id", "params", "expected_input_namespace", "expected_input_name"),
        (
            (
                "google_cloud_storage",
                {
                    "data_path_template": "gs://bucket/path/to/file.txt",
                    "destination_table_name_template": "bq_table",
                },
                "gs://bucket",
                "path/to/file.txt",
            ),
            (
                "amazon_s3",
                {
                    "data_path": "s3://bucket/path/to/file.txt",
                    "destination_table_name_template": "bq_table",
                },
                "s3://bucket",
                "path/to/file.txt",
            ),
            (
                "azure_blob_storage",
                {
                    "storage_account": "account_id",
                    "container": "container_id",
                    "data_path": "/path/to/file.txt",
                    "destination_table_name_template": "bq_table",
                },
                "abfss://container_id@account_id.dfs.core.windows.net",
                "path/to/file.txt",
            ),
        ),
    )
    @mock.patch(
        f"{OPERATOR_MODULE_PATH}.BigQueryDataTransferServiceStartTransferRunsOperator"
        f"._wait_for_transfer_to_be_done"
    )
    @mock.patch(f"{OPERATOR_MODULE_PATH}.BiqQueryDataTransferServiceHook")
    def test_get_openlineage_facets_on_complete_with_blob_storage_sources(
        self, mock_hook, mock_wait, data_source_id, params, expected_input_namespace, expected_input_name
    ):
        mock_hook.return_value.start_manual_transfer_runs.return_value = StartManualTransferRunsResponse(
            runs=[TransferRun(name=RUN_NAME)]
        )
        mock_wait.return_value = {
            "error_status": {"code": 0, "message": "", "details": []},
            "data_source_id": data_source_id,
            "destination_dataset_id": "bq_dataset",
            "params": params,
        }

        op = BigQueryDataTransferServiceStartTransferRunsOperator(
            transfer_config_id=TRANSFER_CONFIG_ID, task_id="id", project_id=PROJECT_ID
        )
        op.execute({"ti": mock.MagicMock(), "task": mock.MagicMock()})
        result = op.get_openlineage_facets_on_complete(None)
        assert not result.run_facets
        assert not result.job_facets
        assert len(result.inputs) == 1
        assert len(result.outputs) == 1
        assert result.inputs[0].namespace == expected_input_namespace
        assert result.inputs[0].name == expected_input_name
        assert result.outputs[0].namespace == "bigquery"
        assert result.outputs[0].name == f"{PROJECT_ID}.bq_dataset.bq_table"

    @pytest.mark.parametrize(
        ("data_source_id", "params", "expected_input_namespace", "expected_input_names"),
        (
            (
                "postgresql",
                {
                    "connector.endpoint.host": "127.0.0.1",
                    "connector.endpoint.port": 5432,
                    "assets": [
                        "db1/sch1/tb1",
                        "db2/sch2/tb2",
                    ],
                },
                "postgres://127.0.0.1:5432",
                ["db1.sch1.tb1", "db2.sch2.tb2"],
            ),
            (
                "oracle",
                {
                    "connector.endpoint.host": "127.0.0.1",
                    "connector.endpoint.port": 1234,
                    "assets": [
                        "db1/sch1/tb1",
                        "db2/sch2/tb2",
                    ],
                },
                "oracle://127.0.0.1:1234",
                ["db1.sch1.tb1", "db2.sch2.tb2"],
            ),
            (
                "mysql",
                {
                    "connector.endpoint.host": "127.0.0.1",
                    "connector.endpoint.port": 3306,
                    "assets": [
                        "db1/tb1",
                        "db2/tb2",
                    ],
                },
                "mysql://127.0.0.1:3306",
                ["db1.tb1", "db2.tb2"],
            ),
        ),
    )
    @mock.patch(
        f"{OPERATOR_MODULE_PATH}.BigQueryDataTransferServiceStartTransferRunsOperator"
        f"._wait_for_transfer_to_be_done"
    )
    @mock.patch(f"{OPERATOR_MODULE_PATH}.BiqQueryDataTransferServiceHook")
    def test_get_openlineage_facets_on_complete_with_sql_sources(
        self, mock_hook, mock_wait, data_source_id, params, expected_input_namespace, expected_input_names
    ):
        mock_hook.return_value.start_manual_transfer_runs.return_value = StartManualTransferRunsResponse(
            runs=[TransferRun(name=RUN_NAME)]
        )
        mock_wait.return_value = {
            "error_status": {"code": 0, "message": "", "details": []},
            "data_source_id": data_source_id,
            "destination_dataset_id": "bq_dataset",
            "params": params,
        }

        op = BigQueryDataTransferServiceStartTransferRunsOperator(
            transfer_config_id=TRANSFER_CONFIG_ID, task_id="id", project_id=PROJECT_ID
        )
        op.execute({"ti": mock.MagicMock(), "task": mock.MagicMock()})
        result = op.get_openlineage_facets_on_complete(None)
        assert not result.run_facets
        assert not result.job_facets
        assert len(result.inputs) == 2
        assert len(result.outputs) == 2
        assert result.inputs[0].namespace == expected_input_namespace
        assert result.inputs[0].name == expected_input_names[0]
        assert result.inputs[1].namespace == expected_input_namespace
        assert result.inputs[1].name == expected_input_names[1]
        assert result.outputs[0].namespace == "bigquery"
        assert result.outputs[0].name == f"{PROJECT_ID}.bq_dataset.{expected_input_names[0].split('.')[-1]}"
        assert result.outputs[1].namespace == "bigquery"
        assert result.outputs[1].name == f"{PROJECT_ID}.bq_dataset.{expected_input_names[1].split('.')[-1]}"

    @mock.patch(
        f"{OPERATOR_MODULE_PATH}.BigQueryDataTransferServiceStartTransferRunsOperator"
        f"._wait_for_transfer_to_be_done"
    )
    @mock.patch(f"{OPERATOR_MODULE_PATH}.BiqQueryDataTransferServiceHook")
    def test_get_openlineage_facets_on_complete_with_scheduled_query(self, mock_hook, mock_wait):
        mock_hook.return_value.start_manual_transfer_runs.return_value = StartManualTransferRunsResponse(
            runs=[TransferRun(name=RUN_NAME)]
        )
        mock_wait.return_value = {
            "error_status": {"code": 0, "message": "", "details": []},
            "data_source_id": "scheduled_query",
            "destination_dataset_id": "bq_dataset",
            "params": {"query": "SELECT a,b,c from x.y.z;", "destination_table_name_template": "bq_table"},
        }

        op = BigQueryDataTransferServiceStartTransferRunsOperator(
            transfer_config_id=TRANSFER_CONFIG_ID, task_id="id", project_id=PROJECT_ID
        )
        op.execute({"ti": mock.MagicMock(), "task": mock.MagicMock()})
        result = op.get_openlineage_facets_on_complete(None)
        assert len(result.job_facets) == 1
        assert result.job_facets["sql"].query == "SELECT a,b,c from x.y.z"
        assert not result.run_facets
        assert len(result.inputs) == 1
        assert len(result.outputs) == 1
        assert result.inputs[0].namespace == "bigquery"
        assert result.inputs[0].name == "x.y.z"
        assert result.outputs[0].namespace == "bigquery"
        assert result.outputs[0].name == f"{PROJECT_ID}.bq_dataset.bq_table"

    @mock.patch(
        f"{OPERATOR_MODULE_PATH}.BigQueryDataTransferServiceStartTransferRunsOperator"
        f"._wait_for_transfer_to_be_done"
    )
    @mock.patch(f"{OPERATOR_MODULE_PATH}.BiqQueryDataTransferServiceHook")
    def test_get_openlineage_facets_on_complete_with_error(self, mock_hook, mock_wait):
        mock_hook.return_value.start_manual_transfer_runs.return_value = StartManualTransferRunsResponse(
            runs=[TransferRun(name=RUN_NAME)]
        )
        mock_wait.return_value = {
            "error_status": {
                "code": 1,
                "message": "Sample message error.",
                "details": [{"@type": "123", "field1": "test1"}, {"@type": "456", "field2": "test2"}],
            },
            "data_source_id": "google_cloud_storage",
            "destination_dataset_id": "bq_dataset",
            "params": {
                "data_path_template": "gs://bucket/path/to/file.txt",
                "destination_table_name_template": "bq_table",
            },
        }

        op = BigQueryDataTransferServiceStartTransferRunsOperator(
            transfer_config_id=TRANSFER_CONFIG_ID, task_id="id", project_id=PROJECT_ID
        )
        op.execute({"ti": mock.MagicMock(), "task": mock.MagicMock()})
        result = op.get_openlineage_facets_on_complete(None)
        assert not result.job_facets
        assert len(result.run_facets) == 1
        assert result.run_facets["errorMessage"].message == "Sample message error."
        assert (
            result.run_facets["errorMessage"].stackTrace
            == "[{'@type': '123', 'field1': 'test1'}, {'@type': '456', 'field2': 'test2'}]"
        )
        assert len(result.inputs) == 1
        assert len(result.outputs) == 1
        assert result.inputs[0].namespace == "gs://bucket"
        assert result.inputs[0].name == "path/to/file.txt"
        assert result.outputs[0].namespace == "bigquery"
        assert result.outputs[0].name == f"{PROJECT_ID}.bq_dataset.bq_table"

    @mock.patch(f"{OPERATOR_MODULE_PATH}.BiqQueryDataTransferServiceHook")
    @mock.patch(f"{OPERATOR_MODULE_PATH}.BigQueryDataTransferServiceStartTransferRunsOperator.defer")
    def test_get_openlineage_facets_on_complete_deferred(self, mock_defer, mock_hook):
        mock_hook.return_value.start_manual_transfer_runs.return_value = StartManualTransferRunsResponse(
            runs=[TransferRun(name=RUN_NAME)]
        )
        mock_hook.return_value.get_transfer_run.return_value = TransferRun(
            {
                "error_status": {"code": 0, "message": "", "details": []},
                "data_source_id": "google_cloud_storage",
                "destination_dataset_id": "bq_dataset",
                "params": {
                    "data_path_template": "gs://bucket/path/to/file.txt",
                    "destination_table_name_template": "bq_table",
                },
            }
        )

        op = BigQueryDataTransferServiceStartTransferRunsOperator(
            transfer_config_id=TRANSFER_CONFIG_ID, task_id="id", project_id=PROJECT_ID, deferrable=True
        )
        op.execute({"ti": mock.MagicMock(), "task": mock.MagicMock()})
        # `defer` is mocked so it will not call the `execute_completed`, so we do it manually.
        op.execute_completed(
            mock.MagicMock(), {"status": "done", "run_id": 123, "config_id": 321, "message": "msg"}
        )
        result = op.get_openlineage_facets_on_complete(None)
        assert not result.job_facets
        assert not result.run_facets
        assert len(result.inputs) == 1
        assert len(result.outputs) == 1
        assert result.inputs[0].namespace == "gs://bucket"
        assert result.inputs[0].name == "path/to/file.txt"
        assert result.outputs[0].namespace == "bigquery"
        assert result.outputs[0].name == f"{PROJECT_ID}.bq_dataset.bq_table"
